# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Access project's data and documentation files.

All access to data and documentation files MUST be made through the functions
of this modules to ensure access across different distribution schemes:

- Installing from source or from wheel
- Installing package as a zip (through the use of pkg_resources)
- Linux packaging willing to install data files (and doc files) in
  alternative folders. In this case, this file must be patched.
- Frozen fat binary application using silx (frozen with cx_Freeze or py2app).
  This needs special care for the resource files in the setup:

  - With cx_Freeze, add silx/resources to include_files:

    .. code-block:: python

       import silx.resources
       silx_include_files = (os.path.dirname(silx.resources.__file__),
                             os.path.join('silx', 'resources'))
       setup(...
             options={'build_exe': {'include_files': [silx_include_files]}}
             )

  - With py2app, add silx in the packages list of the py2app options:

    .. code-block:: python

       setup(...
             options={'py2app': {'packages': ['silx']}}
             )
"""

__authors__ = ["V.A. Sole", "Thomas Vincent", "J. Kieffer"]
__license__ = "MIT"
__date__ = "08/03/2019"


import os
import sys
import logging
import importlib


logger = logging.getLogger(__name__)


# pkg_resources is useful when this package is stored in a zip
# When pkg_resources is not available, the resources dir defaults to the
# directory containing this module.
try:
    import pkg_resources
except ImportError:
    pkg_resources = None


# For packaging purpose, patch this variable to use an alternative directory
# E.g., replace with _RESOURCES_DIR = '/usr/share/silx/data'
_RESOURCES_DIR = None

# For packaging purpose, patch this variable to use an alternative directory
# E.g., replace with _RESOURCES_DIR = '/usr/share/silx/doc'
# Not in use, uncomment when functionality is needed
# _RESOURCES_DOC_DIR = None

# cx_Freeze frozen support
# See http://cx-freeze.readthedocs.io/en/latest/faq.html#using-data-files
if getattr(sys, "frozen", False):
    # Running in a frozen application:
    # We expect resources to be located either in a silx/resources/ dir
    # relative to the executable or within this package.
    _dir = os.path.join(os.path.dirname(sys.executable), "silx", "resources")
    if os.path.isdir(_dir):
        _RESOURCES_DIR = _dir


class _ResourceDirectory(object):
    """Store a source of resources"""

    def __init__(self, package_name, package_path=None, forced_path=None):
        if forced_path is None:
            if package_path is None:
                if pkg_resources is None:
                    # In this case we have to compute the package path
                    # Else it will not be used
                    module = importlib.import_module(package_name)
                    package_path = os.path.abspath(os.path.dirname(module.__file__))
        self.package_name = package_name
        self.package_path = package_path
        self.forced_path = forced_path


_SILX_DIRECTORY = _ResourceDirectory(
    package_name=__name__,
    package_path=os.path.abspath(os.path.dirname(__file__)),
    forced_path=_RESOURCES_DIR,
)

_RESOURCE_DIRECTORIES = {}
_RESOURCE_DIRECTORIES["silx"] = _SILX_DIRECTORY


def register_resource_directory(name, package_name, forced_path=None):
    """Register another resource directory to the available list.

    By default only the directory "silx" is available.

    .. versionadded:: 0.6

    :param str name: Name of the resource directory. It is used on the resource
        name to specify the resource directory to use. The resource
        "silx:foo.png" will use the "silx" resource directory.
    :param str package_name: Python name of the package containing resources.
        For example "silx.resources".
    :param str forced_path: Path containing the resources. If specified
        `pkg_resources` nor `package_name` will be used
        For example "silx.resources".
    :raises ValueError: If the resource directory name already exists.
    """
    if name in _RESOURCE_DIRECTORIES:
        raise ValueError("Resource directory name %s already exists" % name)
    resource_directory = _ResourceDirectory(
        package_name=package_name, forced_path=forced_path
    )
    _RESOURCE_DIRECTORIES[name] = resource_directory


def list_dir(resource):
    """List the content of a resource directory.

    Result are not prefixed by the resource name.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param str resource: Name of the resource directory to list
    :return: list of name contained in the directory
    :rtype: List
    """
    resource_directory, resource_name = _get_package_and_resource(resource)

    if resource_directory.forced_path is not None:
        # if set, use this directory
        path = resource_filename(resource)
        return os.listdir(path)
    elif pkg_resources is None:
        # Fallback if pkg_resources is not available
        path = resource_filename(resource)
        return os.listdir(path)
    else:
        # Preferred way to get resources as it supports zipfile package
        package_name = resource_directory.package_name
        return pkg_resources.resource_listdir(package_name, resource_name)


def is_dir(resource):
    """True is the resource is a resource directory.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param str resource: Name of the resource
    :rtype: bool
    """
    path = resource_filename(resource)
    return os.path.isdir(path)


def exists(resource):
    """True is the resource exists.

    :param str resource: Name of the resource
    :rtype: bool
    """
    path = resource_filename(resource)
    return os.path.exists(path)


def _get_package_and_resource(resource, default_directory=None):
    """
    Return the resource directory class and a cleaned resource name without
    prefix.

    :param str: resource: Name of the resource with resource prefix.
    :param str default_directory: If the resource is not prefixed, the resource
        will be searched on this default directory of the silx resource
        directory.
    :rtype: tuple(_ResourceDirectory, str)
    :raises ValueError: If the resource name uses an unregistred resource
        directory name
    """
    if ":" in resource:
        prefix, resource = resource.split(":", 1)
    else:
        prefix = "silx"
        if default_directory is not None:
            resource = os.path.join(default_directory, resource)
    if prefix not in _RESOURCE_DIRECTORIES:
        raise ValueError("Resource '%s' uses an unregistred prefix", resource)
    resource_directory = _RESOURCE_DIRECTORIES[prefix]
    return resource_directory, resource


def resource_filename(resource):
    """Return filename corresponding to resource.

    The existence of the resource is not checked.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param str resource: Resource path relative to resource directory
                         using '/' path separator. It can be either a file or
                         a directory.
    :raises ValueError: If the resource name uses an unregistred resource
        directory name
    :return: Absolute resource path in the file system
    :rtype: str
    """
    return _resource_filename(resource, default_directory=None)


def _resource_filename(resource, default_directory=None):
    """Return filename corresponding to resource.

    The existence of the resource is not checked.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param str resource: Resource path relative to resource directory
                         using '/' path separator. It can be either a file or
                         a directory.
    :param str default_directory: If the resource is not prefixed, the resource
        will be searched on this default directory of the silx resource
        directory. It should only be used internally by silx.
    :return: Absolute resource path in the file system
    :rtype: str
    """
    resource_directory, resource_name = _get_package_and_resource(
        resource, default_directory=default_directory
    )

    if resource_directory.forced_path is not None:
        # if set, use this directory
        base_dir = resource_directory.forced_path
        resource_path = os.path.join(base_dir, *resource_name.split("/"))
        return resource_path
    elif pkg_resources is None:
        # Fallback if pkg_resources is not available
        base_dir = resource_directory.package_path
        resource_path = os.path.join(base_dir, *resource_name.split("/"))
        return resource_path
    else:
        # Preferred way to get resources as it supports zipfile package
        package_name = resource_directory.package_name
        return pkg_resources.resource_filename(package_name, resource_name)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/07/2019"


from silx.gui import qt
from Orange.widgets import gui
import logging

_logger = logging.getLogger(__file__)


class _ProcessForOrangeMixIn(object):
    """
    Group processing and progress display in a common class for xasObject
    process.

    If this process own a widget to display the xas object then this one should
    be named '_window'
    """

    def __init__(self):
        # progress
        self._progress = gui.ProgressBar(self, 100)
        """progress bar"""
        self.__processingThread = None
        """Thread for processing"""
        # progress
        self._progress = gui.ProgressBar(self, 100)

    def _startProcess(self):
        if hasattr(self, "_window"):
            self._window.setEnabled(False)
        self._progress.widget.progressBarInit()

    def _endProcess(self, xas_obj):
        if hasattr(self, "_window"):
            self._window.setEnabled(True)
        if self._callback_finish:
            try:
                self.getProcessingThread()._process_obj._advancement.sigProgress.disconnect(
                    self._setProgressValue
                )
            except ... as e:
                _logger.error(str(e))

            self.getProcessingThread().finished.disconnect(self._callback_finish)
            self._callback_finish = None
        if xas_obj is None:
            return
        else:
            if hasattr(self, "_window") and hasattr(self._window, "setXASObj"):
                self._window.setXASObj(xas_obj=xas_obj)
            elif hasattr(self, "_window") and hasattr(self._window, "xasObjViewer"):
                if hasattr(self._window.xasObjViewer, "setXASObj"):
                    self._window.xasObjViewer.setXASObj(xas_obj=xas_obj)
            # emit signal for the plot
            self.Outputs.res_xas_obj.send(xas_obj)

    def _canProcess(self):
        return (
            self.__processingThread is None or not self.__processingThread.isRunning()
        )

    def getProcessingThread(self):
        if self.__processingThread is None:
            self.__processingThread = ProcessQThread(parent=self)
        return self.__processingThread

    def _setProgressValue(self, value):
        self._progress.widget.progressBarSet(value)


class ProcessRunnable(qt.QRunnable):
    """
    qt Runnable for standard process.
    process function should take as input(spectrum, configuration, overwrite)

    :param function pointer fct: process function
    :param :class:`.Spectrum`: spectrum to process
    :param dict configuration: configuration of the process
    :param function pointer callback: optional callback to execute at the end of
                                      the run. Should take no parameter
    """

    def __init__(self, fct, spectrum, configuration, callback=None):
        qt.QRunnable.__init__(self)
        self._spectrum = spectrum
        self._configuration = configuration
        self._callback = callback
        self._function = fct

    def run(self):
        try:
            self._configuration, self._spectrum = self._function(
                spectrum=self._spectrum,
                configuration=self._configuration,
                overwrite=True,
            )
        except (KeyError, ValueError) as e:
            _logger.error(e)
        if self._callback:
            self._callback()


class ProcessQThread(qt.QThread):
    """
    Thread dedicated to process execution.
    """

    def __init__(self, parent=None):
        qt.QThread.__init__(self, parent)

    def init(self, xas_obj, process_obj):
        """
        Initialize the thread for processing xas_obj from proces_obj

        :param :class:`.XASObject` xas_obj: object to process
        :param :class:`.Process` process_obj: object to process xas_obj
        """
        self._xas_obj = xas_obj
        self._process_obj = process_obj

    def run(self):
        self._xas_obj = self._process_obj.process(self._xas_obj)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/30/2019"

from silx.gui import qt
from est.gui.larch.utils import _OptionalQIntSpinBox


class _MXFTFParameters(qt.QWidget):
    """
    Widget for setting the configuration of the larch 'xftf' process
    """

    sigChanged = qt.Signal()
    """Signal emitted when parameters changed"""

    _VALID_WINDOWS = ("kaiser", "hanning", "parzen", "welch", "gaussian", "sine")

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QFormLayout())

        # kmin
        self._kminSB = qt.QSpinBox(parent=self)
        self._kminSB.setValue(0)
        self._kminSB.setRange(0, 1000)
        self._kminSB.setToolTip("starting k for FT Window")
        self.layout().addRow(qt.QLabel("k min", parent=self), self._kminSB)
        # kmax
        self._kmaxSB = qt.QSpinBox(parent=self)
        self._kmaxSB.setRange(1, 2000)
        self._kmaxSB.setValue(20)
        self._kmaxSB.setToolTip("ending k for FT Window")
        self.layout().addRow(qt.QLabel("k max", parent=self), self._kmaxSB)
        # kweight
        self._kWeightSB = qt.QSpinBox(parent=self)
        self._kWeightSB.setValue(0)
        self._kWeightSB.setToolTip("exponent for weighting spectra by " "k**kweight")
        self.layout().addRow(qt.QLabel("k weight", parent=self), self._kWeightSB)
        # dk1
        self._dk1SB = qt.QSpinBox(parent=self)
        self._dk1SB.setValue(1)
        self._dk1SB.setToolTip("tapering parameter for FT Window")
        self.layout().addRow(qt.QLabel("dk", parent=self), self._dk1SB)
        # dk2
        self._dk2SB = _OptionalQIntSpinBox(parent=self)
        self._dk2SB.setValue(None)
        self._dk2SB.setToolTip("second tapering parameter for FT Window")
        self.layout().addRow(qt.QLabel("dk2", parent=self), self._dk2SB)
        # with phase
        self._withPhaseCB = qt.QCheckBox(parent=self)
        self._withPhaseCB.setChecked(False)
        self._withPhaseCB.setToolTip(
            "output the phase as well as magnitude, " "real, imag"
        )
        self.layout().addRow(qt.QLabel("phase", parent=self), self._withPhaseCB)
        # window
        self._windowCB = qt.QComboBox(parent=self)
        for winName in self._VALID_WINDOWS:
            self._windowCB.addItem(winName)
        index = self._windowCB.findText("kaiser")
        assert index >= 0
        self._windowCB.setCurrentIndex(index)
        self.layout().addRow(qt.QLabel("window", parent=self), self._windowCB)
        # rmax_out
        self._rmax_outSB = qt.QDoubleSpinBox(parent=self)
        self._rmax_outSB.setValue(10.0)
        self._rmax_outSB.setRange(0, 999999)
        self._rmax_outSB.setToolTip("highest R for output data")
        self.layout().addRow(
            qt.QLabel("highest R (Ang)", parent=self), self._rmax_outSB
        )
        # nfft
        self._nfftSB = qt.QSpinBox(parent=self)
        self._nfftSB.setMaximum(999999)
        self._nfftSB.setValue(2048)
        self._nfftSB.setToolTip("value to use for N_fft")
        self.layout().addRow(qt.QLabel("N fft", parent=self), self._nfftSB)
        # kstep
        self._kstepSB = qt.QDoubleSpinBox(parent=self)
        self._kstepSB.setToolTip("value to use for delta_k")
        self._kstepSB.setValue(0.05)
        self._kstepSB.setSingleStep(0.05)
        self.layout().addRow(qt.QLabel("k step", parent=self), self._kstepSB)

        # expose API
        self.getKMin = self._kminSB.value
        self.setKMin = self._kminSB.setValue
        self.getKMax = self._kmaxSB.value
        self.setKMax = self._kmaxSB.setValue
        self.getKWeight = self._kWeightSB.value
        self.setKWeight = self._kWeightSB.setValue
        self.setDK1 = self._dk1SB.setValue
        self.getDK1 = self._dk1SB.value
        self.setDK2 = self._dk2SB.setValue
        self.getDK2 = self._dk2SB.getValue
        self.isUsingPhase = self._withPhaseCB.isChecked
        self.setUsingPhase = self._withPhaseCB.setChecked
        self.getWindow = self._windowCB.currentText
        self.setRMaxOut = self._rmax_outSB.setValue
        self.getRMaxOut = self._rmax_outSB.value
        self.getNfft = self._nfftSB.value
        self.setNfft = self._nfftSB.setValue
        self.getKStep = self._kstepSB.value
        self.setKStep = self._kstepSB.setValue

        # connect signal / Slot
        self._kminSB.valueChanged.connect(self._valueChanged)
        self._kmaxSB.valueChanged.connect(self._valueChanged)
        self._kWeightSB.valueChanged.connect(self._valueChanged)
        self._dk1SB.valueChanged.connect(self._valueChanged)
        self._dk2SB.sigChanged.connect(self._valueChanged)
        self._withPhaseCB.toggled.connect(self._valueChanged)
        self._windowCB.currentIndexChanged.connect(self._valueChanged)
        self._rmax_outSB.valueChanged.connect(self._valueChanged)
        self._nfftSB.valueChanged.connect(self._valueChanged)
        self._kstepSB.valueChanged.connect(self._valueChanged)

        # avoid several emission of the sigChanged signal
        self._lastParameters = None

    def setWindow(self, window):
        assert window in self._VALID_WINDOWS
        index = self._windowCB.findText(window)
        assert index >= 0
        self._windowCB.setCurrentIndex(index)

    def _valueChanged(self, *arg, **kwargs):
        currentParameters = self.getParameters()
        if currentParameters != self._lastParameters:
            self._lastParameters = currentParameters
            self.sigChanged.emit()

    def getParameters(self):
        return {
            "kmin": self.getKMin(),
            "kmax": self.getKMax(),
            "kweight": self.getKWeight(),
            "dk": self.getDK1(),
            "dk2": self.getDK2(),
            "with_phase": self.isUsingPhase(),
            "window": self.getWindow(),
            "rmax_out": self.getRMaxOut(),
            "nfft": self.getNfft(),
            "kstep": self.getKStep(),
        }

    def setParameters(self, parameters):
        assert isinstance(parameters, dict)
        for key, value in parameters.items():
            if key == "kmin":
                self.setKMin(value)
            elif key == "kmax":
                self.setKMax(value)
            elif key == "kweight":
                self.setKWeight(value)
            elif key == "dk":
                self.setDK1(value)
            elif key == "dk2":
                self.setDK2(value)
            elif key == "with_phase":
                self.setUsingPhase(value)
            elif key == "window":
                self.setWindow(value)
            elif key == "rmax_out":
                self.setRMaxOut(value)
            elif key == "nfft":
                self.setNfft(value)
            elif key == "kstep":
                self.setKStep(value)


if __name__ == "__main__":
    app = qt.QApplication([])
    widget = _MXFTFParameters()
    widget.show()
    app.exec_()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""module for process base class"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/08/2019"


from .progress import Progress
from est.core.types import XASObject
from ..utils import extract_properties_from_dict
import logging
from collections import namedtuple
from typing import Iterable

_logger = logging.getLogger(__name__)


_input_desc = namedtuple("_input_desc", ["name", "type", "handler", "doc"])

_output_desc = namedtuple("_output_desc", ["name", "type", "doc"])


class _NexusSpectrumDef:
    """Util function to define a Nexus plot"""

    def __init__(
        self,
        signal,
        axes,
        auxiliary_signals,
        silx_style=None,
        title=None,
        title_latex=None,
    ):
        self.__signal = None
        self.__axes = None
        self.__auxiliary_signals = None
        self.__title = None
        self.__title_latex = None
        self.__silx_style = silx_style

        self.signal = signal
        self.axes = axes
        self.auxiliary_signals = auxiliary_signals
        self.title = title

    @property
    def signal(self):
        return self.__signal

    @signal.setter
    def signal(self, signal):
        if not isinstance(signal, str):
            raise TypeError("signal should be an instance of str")
        else:
            self.__signal = signal

    @property
    def axes(self):
        return self.__axes

    @axes.setter
    def axes(self, axes):
        if not isinstance(axes, tuple):
            raise TypeError("axes should be an instance of tuple")
        else:
            self.__axes = axes

    @property
    def auxiliary_signals(self):
        return self.__auxiliary_signals

    @auxiliary_signals.setter
    def auxiliary_signals(self, auxiliary_signals):
        if not isinstance(auxiliary_signals, (tuple, type(None))):
            raise TypeError("auxiliary_signals should be an instance of tuple")
        else:
            self.__auxiliary_signals = auxiliary_signals

    @property
    def silx_style(self):
        if self.__silx_style is None:
            return {}
        else:
            return self.__silx_style

    @property
    def title(self):
        return self.__title

    @title.setter
    def title(self, title):
        if not isinstance(title, (str, type(None))):
            raise TypeError("`title` should be an instance of str or None")
        else:
            self.__title = title

    @property
    def title_latex(self):
        return self.__title_latex

    @title.setter
    def title_latex(self, title):
        if not isinstance(title, (str, type(None))):
            raise TypeError("`title` should be an instance of str or None")
        else:
            self.__title_latex = title


class _NexusDatasetDef:
    """Util function to define a Nexus plot"""

    def __init__(self, name: str, units=None, units_latex=None):
        self.__name = None
        self.__units = None
        self.__units_latex = None

        assert isinstance(name, str)

        self.name = name
        self.units = units
        self.units_latex = units_latex

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError("name should be an instance of str")
        else:
            self.__name = name

    @property
    def units(self):
        return self.__units

    @units.setter
    def units(self, units):
        if not isinstance(units, (str, type(None))):
            raise TypeError("units should be an instance of str")
        else:
            self.__units = units

    @property
    def units_latex(self):
        return self.__units_latex

    @units_latex.setter
    def units_latex(self, units):
        if not isinstance(units, (str, type(None))):
            raise TypeError("units should be an instance of str")
        else:
            self.__units_latex = units

    @property
    def attrs(self):
        attrs = {}
        if self.units is not None:
            attrs.update({"units": self.units})
        if self.units_latex is not None:
            attrs.update({"units_latex": self.units_latex})
        return attrs

    def __str__(self):
        return self.name


class Process(object):
    def __init__(self, name):
        assert type(name) is str
        self._name = name
        self._advancement = Progress(name=name)
        self.__stop = False
        """flag to notice when a end of process is required"""
        self._settings = {}
        # configuration
        self._callbacks = []

    @property
    def name(self) -> str:
        return self._name

    def stop(self):
        self.__stop = True

    @property
    def advancement(self):
        return self._advancement

    @advancement.setter
    def advancement(self, advancement):
        assert isinstance(advancement, Progress)
        self._advancement = advancement

    @property
    def callbacks(self):
        return self._callbacks

    @staticmethod
    def getXasObject(xas_obj) -> XASObject:
        if isinstance(xas_obj, dict):
            _xas_obj = XASObject.from_dict(xas_obj)
        else:
            _xas_obj = xas_obj
        assert isinstance(_xas_obj, XASObject)
        if _xas_obj.n_spectrum > 0:
            _xas_obj.spectra.check_validity()
        assert isinstance(_xas_obj, XASObject)
        return _xas_obj

    def program_name(self) -> str:
        """Name of the program used for this processing"""
        raise NotImplementedError("Base class")

    def program_version(self) -> str:
        """version of the program used for this processing"""
        raise NotImplementedError("Base class")

    def definition(self) -> str:
        """definition of the process"""
        raise NotImplementedError("Base class")

    def getConfiguration(self) -> dict:
        """

        :return: configuration of the process
        :rtype: dict
        """
        if len(self._settings) > 0:
            return self._settings
        else:
            return None

    def setConfiguration(self, configuration: dict):
        # filter configuration from orange widgets
        if "__version__" in configuration:
            del configuration["__version__"]
        if "savedWidgetGeometry" in configuration:
            del configuration["savedWidgetGeometry"]
        if "savedWidgetGeometry" in configuration:
            del configuration["savedWidgetGeometry"]
        if "controlAreaVisible" in configuration:
            del configuration["controlAreaVisible"]

        self._settings = configuration

    def register_process(
        self, xas_obj: XASObject, data_keys: Iterable, plots: Iterable = tuple()
    ):
        """
        Store the current process in the linked h5 file if any,
        output data stored will be the one defined by the data_keys

        :param xas_obj: object for which we want to save the treatment
        :type: :class:`.XASObject`
        :param tuple data_keys: keys of the id to save
        :param plots:
        """
        if xas_obj.has_linked_file():
            _data = {}
            for data_info in data_keys:
                # data_info can be a string or an instance of _NexusDatasetDef
                key = str(data_info)
                relative_to = None
                use = "map_to"
                if key in (
                    "flat",
                    "fpp",
                    "f2",
                    "dmude",
                    "norm",
                    "norm_area",
                    "post_edge",
                    "bkg",
                    "energy",
                    "mback_mu",
                    "norm_mback",
                    "Energy",
                    "Mu",
                    "mu",
                    "NormalizedEnergy",
                    "NormalizedMu",
                    "NormalizedSignal",
                    "EXAFSKValues",
                    "EXAFSSignal",
                    "mu_ref",
                    "I0",
                    "I1",
                    "I2",
                ):
                    relative_to = "energy"
                    use = "map_to"
                elif key in (
                    "noise_savgol",
                    "noise_savgol_energy",
                ):
                    relative_to = "noise_savgol_energy"
                    use = "map_to"
                elif key in ("chir_re", "chir_im", "chir_mag", "r"):
                    relative_to = "r"
                    use = "map_to"
                elif key in ("masked_chir_mag", "masked_r"):
                    relative_to = "masked_r"
                    use = "map_to"
                elif key in ("ft.radius", "ft.intensity", "ft.imaginary"):
                    relative_to = "radius"
                    use = "_list_res_ft"
                elif key in ("masked_chi_weighted_k", "masked_k"):
                    relative_to = "masked_k"
                    use = "map_to"
                elif key in (
                    "chi",
                    "k",
                    "chi_weighted_k",
                ):
                    relative_to = "k"
                    use = "map_to"

                if use == "map_to":
                    if key == "norm_area":
                        continue
                    _data[key] = xas_obj.spectra.map_to(
                        data_info=key,
                        relative_to=relative_to,
                    )

                    # if we can display the result as a numpy.array 3d
                    try:
                        _data[key] = xas_obj.spectra.map_to(
                            data_info=key,
                            relative_to=relative_to,
                        )
                    except KeyError:
                        mess = "%s: unable to store %s, parameter not found" % (
                            self.name,
                            key,
                        )
                        _logger.error(mess)

                elif use == "_list_res_ft":
                    # for ft we are force to have a dictionary because it
                    # processing can end with a nan or with a numpy array.
                    # combining both would make us end up with a ndarray of
                    # object type
                    key_ = key.split(".")[-1]
                    res = {}
                    for spectrum in xas_obj.spectra:
                        x, y = spectrum.x, spectrum.y
                        res["_".join((str(x), str(y)))] = getattr(spectrum.ft, key_)
                    _data[key] = res
                else:
                    raise ValueError()
                if isinstance(data_info, _NexusDatasetDef):
                    for aname, avalue in data_info.attrs.items():
                        _data[(key, aname)] = avalue
            xas_obj.register_processing(process=self, results=_data, plots=plots)

    def addCallback(self, callback):
        self._callbacks.append(callback)

    def update_properties(self, properties):
        if properties is None:
            return
        if isinstance(properties, str):
            properties = extract_properties_from_dict(properties)
        self._settings.update(properties)



settings_updater = {
    # larch processing
    "est.core.process.larch.autobk.Larch_autobk": "set_autobk_params",
    "est.core.process.larch.mback.Larch_mback": "set_mback_params",
    "est.core.process.larch.mback_norm.Larch_mback_norm": "set_mback_norm_params",
    "est.core.process.larch.pre_edge.Larch_pre_edge": "set_pre_edge_params",
    "est.core.process.larch.xftf.Larch_xftf": "set_xftf_params",
    # utils
    "est.core.process.io.DumpXasObject": "set_output_params",
    "est.core.process.noise.NoiseProcess": "set_noise_params",
    "est.core.process.energyroi.EnergyROIProcess": "set_energyroi_params",
}



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/12/2020"


from silx.io.url import DataUrl
import est.io
import copy
import numpy
import logging
import h5py
import tempfile
import os
import shutil
import pint
from est.units import ur
from est.units import convert_to as convert_unit_to
from silx.io.utils import h5py_read_dataset
from typing import Union
from .dim import Dim
from .spectra import Spectra
from est.thirdparty.silx.hdf5file import get_data, HDF5File

try:
    import larch
except ImportError:
    _has_larch = False
else:
    _has_larch = True

_logger = logging.getLogger(__name__)


class XASObject(object):
    """Base class of XAS

    :param spectra: absorbed beam as a list of :class:`.Spectrum` or a
                    numpy.ndarray. If is a numpy array:
                        * dim0: channel,
                        * dim1: Y,
                        * dim2: X
    :type: Union[numpy.ndarray, list]
    :param energy: beam energy
    :type: numpy.ndarray of one dimension
    :param dict configuration: configuration of the different process
    :param int dim1: first dimension of the spectra
    :param int dim2: second dimension of the spectra
    :param str name: name of the object. Will be used for the hdf5 entry
    :param bool keep_process_flow: if True then will keep the trace of the set
                                   of process applied to the XASObject into a
                                   hdf5 file.
                                   This is also used for the
    :param spectra_url: path to the spectra data if any. Used when serializing
                        the XASObject. Won't read it from it.
    :type: Union[None,str]
    :param spectra_url_dimensions: dimensions of the stored spectra. WARNING:
                                   this is different of the spectra dimension
                                   which should be given in the
                                   `DEFAULT_DIMENSIONS` order (Channel, Y, X)
    :type: Union[tuple,None]
    :param energy_url: path to the energy / channel  data if any. Used when
                       serializing the XASObject. Won't read it from it.
    :type: Union[None,str]
    """

    DEFAULT_DIMENSIONS = (Dim.DIM_2, Dim.DIM_1, Dim.DIM_0)

    def __init__(
        self,
        spectra=None,
        energy=None,
        configuration: Union[dict, None] = None,
        dim1: Union[int, None] = None,
        dim2: Union[int, None] = None,
        name: str = "scan1",
        keep_process_flow: bool = True,
        spectra_url: Union[DataUrl, None] = None,
        spectra_url_dims=None,
        energy_url: Union[DataUrl, None] = None,
        check_energy: bool = True,
    ):
        if isinstance(energy, numpy.ndarray) and not isinstance(
            energy, pint.quantity.Quantity
        ):
            # automatic energy unit guess. convert energy provided in keV to eV
            # works because EXAFS data are never longer than 2 keV and a low
            # energy XANES is never shorter than 3 eV. See issue 30
            if (energy.max() - energy.min()) < 3.0:
                energy = energy * ur.keV
            else:
                energy = energy * ur.eV

        if energy is not None:
            energy = energy.m_as(ur.eV)

        self.__channels = None
        self.__spectra = Spectra(energy=energy)
        self.__dim1 = None
        self.__dim2 = None
        self.__processing_index = 0
        self.__h5_file = None
        self.__entry_name = name
        self.__spectra_url = spectra_url
        self.__spectra_url_dims = spectra_url_dims
        self.__energy_url = energy_url
        self.spectra = (energy, spectra, dim1, dim2)
        self.configuration = configuration
        if keep_process_flow is True:
            self.__h5_file = os.path.join(
                tempfile.mkdtemp(), "_".join((name, "_flow.h5"))
            )
            self.link_to_h5(self.__h5_file)

    def attach_3d_array(self, attr_name, attr_value):
        """
        Attach to each attribute the provided value

        :param attr_name:
        :param attr_value:
        :return:
        """
        if isinstance(attr_value, DataUrl):
            data = get_data(attr_value)
        else:
            data = attr_value
        if not isinstance(data, numpy.ndarray):
            raise TypeError(
                "`attr_value` should be an instance of DataUrl "
                "referencing a numpy array or a numpy array not "
                "{}".format(type(data))
            )
        if data.ndim > 2:
            for spectrum, d in zip(self.spectra.data.flat, data):
                setattr(spectrum, attr_name, d)
        else:
            for spectrum in self.spectra.data.flat:
                setattr(spectrum, attr_name, data)

    @property
    def entry(self) -> str:
        return self.__entry_name

    @property
    def spectra(self):
        return self.__spectra

    @property
    def spectra_url(self) -> Union[None, DataUrl]:
        """Url from where the spectra is available.
        Used for object serialization"""
        return self.__spectra_url

    @spectra_url.setter
    def spectra_url(self, url: DataUrl):
        self.__spectra_url = url

    @property
    def spectra_url_dims(self) -> tuple:
        """used to interpret the spectra_url if any"""
        return self.__spectra_url_dims

    @spectra_url_dims.setter
    def spectra_url_dims(self, dims: tuple):
        self.__spectra_url_dims = dims

    @property
    def energy_url(self) -> Union[DataUrl, None]:
        """Url from where the energy is available.
        Used for object serialization"""
        return self.__energy_url

    @energy_url.setter
    def energy_url(self, url: DataUrl):
        self.__energy_url = url

    @property
    def normalized_energy(self):
        if self.spectra is not None:
            return self.spectra.data.flat[0].normalized_energy

    # TODO: should no take such a tuple in parameter but the different parameter
    # energy, spectra, dim1, dim2
    @spectra.setter
    def spectra(self, energy_spectra: tuple):
        if isinstance(energy_spectra, Spectra):
            self.__spectra = energy_spectra
        else:
            if not len(energy_spectra) == 4:
                raise ValueError(
                    "4 elements expected: energy, spectra, dim1 and dim2. get {} instead".format(
                        len(energy_spectra)
                    )
                )
            energy, spectra, dim1, dim2 = energy_spectra
            self.__spectra = Spectra(energy=energy, spectra=spectra)
            if dim1 is not None and dim2 is not None:
                self.__spectra.reshape((dim1, dim2))

    def get_spectrum(self, dim1_idx: int, dim2_idx: int) -> spectra:
        """Util function to access the spectrum at dim1_idx, dim2_idx"""
        return self.spectra[dim1_idx, dim2_idx]

    @property
    def energy(self) -> numpy.ndarray:
        """energy as a numpy array in eV.
        :note: cannot be with unit because use directly by xraylarch and pymca
        """
        return self.spectra.energy

    @property
    def configuration(self) -> dict:
        return self.__configuration

    @configuration.setter
    def configuration(self, configuration: dict):
        assert configuration is None or isinstance(configuration, dict)
        self.__configuration = configuration or {}

    def _create_saving_pt(self):
        if not self.has_linked_file():
            raise ValueError(
                "there is not where to create a saving pt"
                "(no h5 file linked to the XASObject)"
            )
        else:

            def get_list_spectra():
                res = {}

                for i_spectrum, spectrum in enumerate(self.spectra.data.flat):
                    res[i_spectrum] = spectrum.to_dict()
                return res

            from est.io import write_spectrum_saving_pt

            # save spectra
            entry = "/".join((self.entry, "est_saving_pt", "spectra"))
            for i_spectrum, spectrum in get_list_spectra().items():
                path = "/".join((entry, str(i_spectrum)))
                write_spectrum_saving_pt(
                    h5_file=self.linked_h5_file,
                    entry=path,
                    obj=spectrum,
                    overwrite=True,
                )
            # save channel
            with HDF5File(self.linked_h5_file, "a") as h5f:
                entry = "/".join((self.entry, "est_saving_pt", "channel"))
                if entry in h5f:
                    del h5f[entry]
                h5f[entry] = (
                    "None"
                    if self.energy is None
                    else convert_unit_to(self.energy, ur.eV).m_as(ur.eV)
                )

    def to_dict(self, with_process_details=True) -> dict:
        """convert the XAS object to a dict

        By default made to simply import raw data.

        :param with_process_details: used to embed a list of spectrum with
                                     intermediary result instead of only raw mu.
                                     This is needed especially for the
                                     pushworkflow actors to keep a trace of the
                                     processes.
        :type: bool
        """

        def get_spectra_and_processing():
            if not self.has_linked_file():
                raise ValueError(
                    "To get process details you should have a" "`process` link file"
                )
            else:
                # store the current spectra with processing information
                data_path = "/".join((self.entry, "est_saving_pt", "spectra"))
                return DataUrl(
                    file_path=self.linked_h5_file, data_path=data_path, scheme="est"
                ).path()

        def get_energy():
            if not self.has_linked_file():
                raise ValueError(
                    "To get process details you should have a" "`process` link file"
                )
            else:
                data_path = "/".join((self.entry, "est_saving_pt", "channel"))
                return DataUrl(
                    file_path=self.linked_h5_file, data_path=data_path, scheme="silx"
                ).path()

        self._create_saving_pt()
        spectra_ = get_spectra_and_processing()
        res = {
            "configuration": self.configuration,
            "spectra": spectra_,
            "energy": get_energy(),
            "dim1": self.spectra.shape[0],
            "dim2": self.spectra.shape[1],
        }
        if with_process_details is True:
            res["linked_h5_file"] = self.linked_h5_file
            res["current_processing_index"] = self.__processing_index

        return res

    def absorbed_beam(self) -> numpy.ndarray:
        return self.spectra.map_to(data_info="mu")

    def load_frm_dict(self, ddict: dict):
        """load XAS values from a dict"""
        dimensions = (Dim.DIM_2, Dim.DIM_1, Dim.DIM_0)
        # default yay of storing data
        from est.io import load_data  # avoid cyclic import

        contains_config_spectrum = "configuration" in ddict or "spectra" in ddict
        """The dict can be on the scheme of the to_dict function, containing
        the spectra and the configuration. Otherwise we consider it is simply
        the spectra"""
        if "configuration" in ddict:
            self.configuration = ddict["configuration"]
        if "spectra" in ddict:
            self.spectra = Spectra.frm_dict(ddict["spectra"], dimensions=dimensions)
        else:
            self.spectra = None
        if "energy" in ddict:
            energy = ddict["energy"]
            if isinstance(energy, str):
                energy_url = DataUrl(path=energy)
                energy = load_data(
                    data_url=energy_url, name="energy", dimensions=dimensions
                )
        else:
            energy = None
        if "dim1" in ddict:
            dim1 = ddict["dim1"]
        else:
            dim1 = None
        if "dim2" in ddict:
            dim2 = ddict["dim2"]
        else:
            dim2 = None
        if "linked_h5_file" in ddict:
            assert "current_processing_index" in ddict
            self.link_to_h5(ddict["linked_h5_file"])
            self.__processing_index = ddict["current_processing_index"]
        self.spectra.reshape((dim1, dim2))
        return self

    @staticmethod
    def from_dict(ddict: dict):
        return XASObject().load_frm_dict(ddict=ddict)

    @staticmethod
    def from_file(
        h5_file,
        entry="scan1",
        spectra_path="data/absorbed_beam",
        energy_path="data/energy",
        configuration_path="configuration",
        dimensions=None,
    ):
        # load only mu and energy from the file
        spectra_url = DataUrl(
            file_path=h5_file, data_path="/".join((entry, spectra_path)), scheme="silx"
        )
        energy_url = DataUrl(
            file_path=h5_file, data_path="/".join((entry, energy_path)), scheme="silx"
        )
        if configuration_path is None:
            config_url = None
        else:
            config_url = DataUrl(
                file_path=h5_file,
                data_path="/".join((entry, configuration_path)),
                scheme="silx",
            )
        from est.io.utils.information import InputInformation

        sp, en, conf = est.io.read_xas(
            InputInformation(
                spectra_url=spectra_url,
                channel_url=energy_url,
                config_url=config_url,
                dimensions=dimensions,
            ),
        )
        return XASObject(spectra=sp, energy=en, configuration=conf)

    def dump(self, h5_file: str):
        """dump the XAS object to a file_path within the Nexus format"""
        from est.core.io import XASWriter  # avoid cyclic import

        writer = XASWriter()
        # dump raw data
        writer.output_file = h5_file
        writer.dump_xas(xas_obj=self, write_process=True)

    def copy(self, create_h5_file=False):
        """

        :param bool create_h5_file: True if we want to duplicate the h5 file
                                 This is the case for orange widget for example
                                 where some reprocessing can append and each
                                 process need to keep a clear history of
                                 processes, with no knowledge of next
                                 processing.
        """
        # To have dedicated h5 file we have to create one new h5 file for
        # for each process. For now there is no way to do it differently
        res = copy.copy(self)
        if create_h5_file is True:
            new_h5_file = os.path.join(
                tempfile.mkdtemp(), "_".join((self.__entry_name, "_flow.h5"))
            )
            if self.has_linked_file() and os.path.exists(self.linked_h5_file):
                shutil.copy(src=self.linked_h5_file, dst=new_h5_file)
            res.link_to_h5(new_h5_file)
        return res

    def __eq__(self, other):
        return (
            isinstance(other, XASObject)
            and (
                (self.energy is None and other.energy is None)
                or numpy.array_equal(self.energy, other.energy)  # noqa
            )  # noqa
            and self.configuration == other.configuration  # noqa
            and self.spectra,
            other.spectra,  # noqa
        )

    @property
    def n_spectrum(self) -> int:
        """return the number of spectra"""
        return len(self.spectra.data.flat)

    def spectra_keys(self) -> tuple:
        """keys contained by the spectrum object (energy, mu, normalizedmu...)"""
        return self.spectra.keys()

    @property
    def linked_h5_file(self):
        return self.__h5_file

    def link_to_h5(self, h5_file: str):
        """
        Associate a .h5 file to the XASObject. This can be used for storing
        process flow.

        :param h5_file:
        :return:
        """
        self.__h5_file = h5_file

    def has_linked_file(self) -> bool:
        return self.__h5_file is not None

    def get_next_processing_index(self) -> int:
        self.__processing_index += 1
        return self.__processing_index

    def register_processing(self, process, results, plots) -> None:
        """
        Register one process for the current xas object. This require to having
        link a h5file to this object

        :param process: process to be registred
        :type: :class:`.Process`
        :param results: result of the processing. If there is more than one
                     result then a dictionary with the key under which result
                     should be saved and a numpy.ndarray
        :type: Union[numpy.ndarray,dict]
        """
        est.io.write_xas_proc(
            self.linked_h5_file,
            entry=self.__entry_name,
            processing_order=self.get_next_processing_index(),
            process=process,
            results=results,
            plots=plots,
            overwrite=True,
        )

    def get_process_flow(self) -> dict:
        """

        :return: the dict of process information
        :rtype: dict
        """
        if not self.linked_h5_file:
            _logger.warning(
                "process flow is store in the linked .h5 file. If"
                "no link is defined then this information is not"
                "stored"
            )
            return {}
        else:
            recognized_process = est.io.get_xasproc(
                self.linked_h5_file, entry=self.__entry_name
            )
            know_process = (
                "pymca_normalization",
                "pymca_exafs",
                "pymca_ft",
                "pymca_k_weight",
                "larch_autobk",
                "larch_mback",
                "larch_mback_norm",
                "larch_pre_edge",
                "larch_xftf",
                "noise",
                "noise_savgol",
                "roi",
                "energy-roi",
            )

            def filter_recognized_process(process_list):
                res = []
                for process_ in process_list:
                    if (
                        "program" in process_.keys()
                        and h5py_read_dataset(process_["program"])
                        in know_process  # noqa
                    ):
                        res.append(process_)
                return res

            recognized_process = filter_recognized_process(recognized_process)

            def get_ordered_process(process_list):
                res = {}
                for process_ in process_list:
                    if "processing_order" not in process_:
                        _logger.warning(
                            "one processing has not processing order: "
                            + process_["program"]  # noqa
                        )
                    else:
                        processing_order = int(process_["processing_order"])
                        res[processing_order] = process_
                return res

            return get_ordered_process(recognized_process)

    def clean_process_flow(self) -> None:
        """
        Remove existing process flow
        """
        if not self.linked_h5_file:
            _logger.warning(
                "process flow is store in the linked .h5 file. If"
                "no link is defined then this information is not"
                "stored"
            )
        else:
            process_flow = self.get_process_flow()
            with HDF5File(self.linked_h5_file, "a") as h5f:
                for index, process_ in process_flow.items():
                    del h5f[process_["_h5py_path"]]

    def copy_process_flow_to(self, h5_file_target: str) -> None:
        """
        copy all the recognized process from self.__h5_file to h5_file_target

        :param str h5_file_target: path to the targeted file. Should be an
                                   existing hdf5 file.
        """
        assert os.path.exists(h5_file_target)
        assert h5py.is_hdf5(h5_file_target)

        flow = self.get_process_flow()
        entry = self.entry
        with HDF5File(self.__h5_file, "a") as source_hdf:
            with HDF5File(h5_file_target, "a") as target_hdf:
                target_entry = target_hdf.require_group(entry)

                def remove_entry_prefix(name):
                    return name.replace("/" + entry + "/", "", 1)

                for process_id, process in flow.items():
                    process_path = process["_h5py_path"]
                    dst_path = remove_entry_prefix(name=process_path)
                    target_entry.copy(source=source_hdf[process_path], dest=dst_path)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"

import functools
import logging
import multiprocessing

import numpy
from PyMca5.PyMcaPhysics.xas.XASClass import XASClass
from est.core.process.process import _input_desc
from est.core.process.process import _output_desc
from est.core.process.process import Process
from est.core.process.process import _NexusDatasetDef
from est.core.types import XASObject, Spectrum

_logger = logging.getLogger(__name__)


def process_spectr_ft(
    spectrum,
    configuration,
    overwrite=True,
    callbacks=None,
    output=None,
    output_dict=None,
):
    """

    :param spectrum: spectrum to process
    :type: :class:`.Spectrum`
    :param configuration: configuration of the pymca normalization
    :type: dict
    :param overwrite: False if we want to return a new Spectrum instance
    :type: bool
    :param callback: callback to execute.
    :param output: list to store the result, needed for pool processing
    :type: multiprocessing.manager.list
    :param output_dict: key is input spectrum, value is index in the output
                        list.
    :type: dict
    :return: processed spectrum
    :rtype: tuple (configuration, spectrum)
    """
    _logger.debug(
        "start fourier transform on spectrum (%s, %s)" % (spectrum.x, spectrum.y)
    )
    pymca_xas = XASClass()
    if spectrum.energy is None or spectrum.mu is None:
        _logger.error(
            "Energy and or Mu is/are not specified, unable to " "compute exafs"
        )
        return None, None

    if configuration is not None:
        pymca_xas.setConfiguration(configuration)
    pymca_xas.setSpectrum(energy=spectrum.energy, mu=spectrum.mu)

    if "EXAFSSignal" not in spectrum:
        _logger.warning(
            "exafs has not been processed yet, unable to process" "fourier transform"
        )
        return None, None

    if "EXAFSNormalized" not in spectrum:
        _logger.warning("ft window need to be defined first")
        return None, None

    cleanMu = spectrum["EXAFSSignal"]
    kValues = spectrum["EXAFSKValues"]

    dataSet = numpy.zeros((cleanMu.size, 2), numpy.float)
    dataSet[:, 0] = kValues
    dataSet[:, 1] = cleanMu

    set2 = dataSet.copy()
    set2[:, 1] = spectrum["EXAFSNormalized"]

    # remove points with k<2
    goodi = (set2[:, 0] >= spectrum["KMin"]) & (set2[:, 0] <= spectrum["KMax"])
    set2 = set2[goodi, :]

    if set2.size == 0:
        ft = {"FTImaginary": numpy.nan, "FTIntensity": numpy.nan, "FTRadius": numpy.nan}
    else:
        ft = pymca_xas.fourierTransform(
            set2[:, 0], set2[:, 1], kMin=spectrum["KMin"], kMax=spectrum["KMax"]
        )
        assert "FTIntensity" in ft
        assert "FTRadius" in ft
        assert ft["FTRadius"] is not None
        assert ft["FTIntensity"] is not None
    if callbacks:
        for callback in callbacks:
            callback()

    if overwrite:
        spectrum_ = spectrum
    else:
        spectrum_ = Spectrum()
        spectrum_.update(spectrum)
    assert spectrum_
    spectrum_.ft = ft

    if output is not None:
        assert output_dict is not None
        output[output_dict[spectrum]] = spectrum_
    return configuration, spectrum_


def pymca_ft(xas_obj):
    """

    :param xas_obj: object containing the configuration and spectra to process
    :type: Union[XASObject, dict]
    :return: spectra dict
    :rtype: dict
    """
    ft_obj = PyMca_ft()
    assert xas_obj is not None
    return ft_obj.process(xas_obj=xas_obj)


_USE_MULTIPROCESSING_POOL = False
# note: we cannot use multiprocessing pool with pypushflow for now.


class PyMca_ft(Process):

    inputs = [
        _input_desc(name="xas_obj", type=XASObject, handler="process", doc=""),
    ]

    outputs = [
        _output_desc(name="xas_obj", type=XASObject, doc=""),
    ]

    def __init__(self):
        Process.__init__(self, name="ft")

    def set_properties(self, properties):
        if "_pymcaSettings" in properties:
            self._settings = properties["_pymcaSettings"]

    def process(self, xas_obj):
        """

        :param xas_obj: object containing the configuration and spectra to process
        :type: Union[XASObject, dict]
        :return: spectra dict
        :rtype: dict
        """
        _xas_obj = self.getXasObject(xas_obj=xas_obj)
        if self._settings:
            _xas_obj.configuration["FT"] = self._settings

        self._advancement.reset(max_=_xas_obj.n_spectrum)
        self._advancement.startProcess()
        self._pool_process(xas_obj=_xas_obj)
        self._advancement.endProcess()
        assert hasattr(_xas_obj.spectra.data.flat[0], "ft")
        assert hasattr(_xas_obj.spectra.data.flat[0].ft, "intensity")
        assert hasattr(_xas_obj.spectra.data.flat[0].ft, "imaginary")
        self.register_process(
            _xas_obj,
            data_keys=(
                _NexusDatasetDef("ft.radius"),
                _NexusDatasetDef("ft.intensity"),
                _NexusDatasetDef("ft.imaginary"),
            ),
        )
        return _xas_obj

    def _pool_process(self, xas_obj):
        assert isinstance(xas_obj, XASObject)
        if not _USE_MULTIPROCESSING_POOL:
            for spectrum in xas_obj.spectra:
                process_spectr_ft(
                    spectrum=spectrum,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=True,
                )
        else:
            from multiprocessing import Manager

            manager = Manager()
            output_dict = {}
            res_list = manager.list()
            for i_spect, spect in enumerate(xas_obj.spectra):
                res_list.append(None)
                output_dict[spect] = i_spect

            with multiprocessing.Pool(1) as p:
                partial_ = functools.partial(
                    process_spectr_ft,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=False,
                    output=res_list,
                    output_dict=output_dict,
                )
                p.map(partial_, xas_obj.spectra)

            # then update local spectrum
            for spectrum, res in zip(xas_obj.spectra, res_list):
                spectrum.update(res)

    def definition(self):
        return "fourier transform"

    def program_version(self):
        import PyMca5

        return PyMca5.version()

    def program_name(self):
        return "pymca_ft"

    __call__ = process


if __name__ == "__main__":
    import sys
    import yaml

    xas_object_yaml_file = sys.argv[1]
    _logger.debug("Load xas object from {}".format(xas_object_yaml_file))
    with open(xas_object_yaml_file, "r") as file:
        ddict = yaml.load(file)["input_data"]
        xas_object = XASObject.from_dict(ddict)
    print("******* do ft ********")
    res_xas_object = pymca_ft(xas_obj=xas_object)
    res_xas_object._create_saving_pt()

    # dump resulting object
    with open(xas_object_yaml_file, "w") as file:
        yaml.dump({"input_data": res_xas_object.to_dict()}, file)

    # dump resulting object into output_normalization file
    with open("output_ft.yaml", "w") as file:
        yaml.dump({"input_data": res_xas_object.to_dict()}, file)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/26/2019"


import os
import shutil
import tempfile
import unittest
import numpy
from est.core.io import XASWriter
from est.core.types import XASObject
import urllib.request
from est.core.types import Spectrum
from ..reprocessing import exec_
from est.core.types import Dim

try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from est.core.process.pymca.exafs import PyMca_exafs
    from est.core.process.pymca.ft import PyMca_ft
    from est.core.process.pymca.k_weight import PyMca_k_weight
    from est.core.process.pymca.normalization import PyMca_normalization
    from est.io.utils.spec import read_spectrum
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
try:
    import larch
except ImportError:
    has_larch = False
else:
    has_larch = True
    from est.core.process.larch.pre_edge import Larch_pre_edge
    from est.core.process.larch.xftf import Larch_xftf
    from est.core.process.larch.autobk import Larch_autobk
    from est.io.utils.larch import read_ascii


@unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
class TestReprocessingPyMca(unittest.TestCase):
    """test reprocessing work for some pymca process"""

    def setUp(self):
        self.output_dir = tempfile.mkdtemp()
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        energy, mu = read_spectrum(
            data_file,
            energy_col_name="Column 1",
            absorption_col_name="Column 2",
        )
        self.spectrum = Spectrum(energy=energy, mu=mu)
        self.xas_obj_ref = XASObject(
            spectra=(self.spectrum,), energy=energy, dim1=1, dim2=1
        )
        self.h5_file = os.path.join(self.output_dir, "output_file.h5")
        out = PyMca_normalization()(xas_obj=self.xas_obj_ref)
        out = PyMca_exafs()(xas_obj=out)
        out = PyMca_k_weight()(xas_obj=out)
        out = PyMca_ft()(xas_obj=out)
        out = PyMca_normalization()(xas_obj=out)

        writer = XASWriter()
        writer.output_file = self.h5_file
        writer(out)
        self.assertTrue(out.spectra.data.flat[0].ft.intensity is not None)
        self.assertEqual(len(out.get_process_flow()), 5)

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def test(self):
        res_xas_obj = exec_(self.h5_file)
        self.assertEqual(res_xas_obj, self.xas_obj_ref)
        numpy.testing.assert_allclose(
            res_xas_obj.spectra.data.flat[0].ft.intensity,
            self.xas_obj_ref.spectra.data.flat[0].ft.intensity,
        )


@unittest.skipIf(has_larch is False, "Larch is not installed")
class TestReprocessingLarch(unittest.TestCase):
    """test reprocessing work for some larch process"""

    def setUp(self):
        self.output_dir = tempfile.mkdtemp()
        xmu_url = "https://raw.githubusercontent.com/xraypy/xraylarch/master/examples/xafs/cu_rt01.xmu"
        self.data_file = os.path.join(self.output_dir, "cu_rt01.xmu")

        with urllib.request.urlopen(xmu_url) as response, open(
            self.data_file, "wb"
        ) as out_file:
            data = response.read()  # a `bytes` object
            out_file.write(data)

        assert os.path.exists(self.data_file)
        energy, mu = read_ascii(self.data_file)
        self.spectrum = Spectrum(energy=energy, mu=mu)
        spectra = (self.spectrum,)
        self.xas_obj_ref = XASObject(spectra=spectra, energy=energy, dim1=1, dim2=1)
        self.h5_file = os.path.join(self.output_dir, "output_file.h5")

        pre_edge_process = Larch_pre_edge()
        pre_edge_process.set_properties({"rbkg": 1.0, "kweight": 2})
        autobk_process = Larch_autobk()
        autobk_process.set_properties(
            {"kmin": 2, "kmax": 16, "dk": 3, "window": "hanning", "kweight": 2}
        )
        xftf_process = Larch_xftf()
        xftf_process.set_properties({"kweight": 2})

        out = pre_edge_process(xas_obj=self.xas_obj_ref)
        out = autobk_process(xas_obj=out)
        out = xftf_process(xas_obj=out)

        writer = XASWriter()
        writer.output_file = self.h5_file
        writer(out)
        assert len(out.get_process_flow()) is 3

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def test(self):
        res_xas_obj = exec_(self.h5_file)
        self.assertEqual(res_xas_obj, self.xas_obj_ref)
        numpy.testing.assert_allclose(
            res_xas_obj.spectra.data.flat[0].chir_mag,
            self.xas_obj_ref.spectra.data.flat[0].chir_mag,
        )


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestReprocessingPyMca, TestReprocessingLarch):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/07/2019"

import functools
import logging

from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget
from Orange.widgets.widget import Input, Output
import Orange.data
from silx.gui import qt

import est.core.process.pymca.k_weight
from orangecontrib.est.process import _ProcessForOrangeMixIn, ProcessRunnable
from est.core.types import XASObject
from orangecontrib.est.progress import QProgress
from orangecontrib.est.utils import Converter

_logger = logging.getLogger(__file__)


class KWeightWindow(qt.QMainWindow):
    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        # k wright widget
        self._k_widget = qt.QWidget(parent=self)
        self._k_widget.setLayout(qt.QHBoxLayout())
        self._k_widget.layout().addWidget(qt.QLabel("k weight"))
        self._k_spin_box = qt.QSpinBox(parent=self)
        self._k_spin_box.setRange(0, 3)
        self._k_widget.layout().addWidget(self._k_spin_box)
        dockWidget = qt.QDockWidget(parent=self)
        dockWidget.setWidget(self._k_widget)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, dockWidget)
        dockWidget.setAllowedAreas(qt.Qt.RightDockWidgetArea | qt.Qt.LeftDockWidgetArea)
        dockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)

        self.setWindowFlags(qt.Qt.Widget)


class KWeightOW(_ProcessForOrangeMixIn, OWWidget):
    """
    Widget used for signal extraction
    """

    name = "k weight"
    id = "orange.widgets.est.pymca.k_weight"
    description = "Progress k weight"
    icon = "icons/k_weight.png"
    priority = 2
    category = "esrfWidgets"
    keywords = ["spectroscopy", "signal", "k", "weight"]

    want_main_area = True
    resizing_enabled = True
    allows_cycle = False

    process_function = est.core.process.pymca.k_weight.PyMca_k_weight

    _kWeightSetting = Setting(int(3))
    """Store the configuration of the PyMca XASClass"""

    class Inputs:
        xas_obj = Input("xas_obj", XASObject, default=True)
        # simple compatibility for some Orange widget and especialy the
        # 'spectroscopy add-on'
        data_table = Input("Data", Orange.data.Table)

    class Outputs:
        res_xas_obj = Output("xas_obj", XASObject)
        # by default we want to avoid sending 'Orange.data.Table' to avoid
        # loosing the XASObject flow process and results.

    def __init__(self):
        super().__init__()
        self._latest_xas_obj = None
        layout = gui.vBox(self.mainArea, "k weight").layout()
        self._window = KWeightWindow(parent=self)
        layout.addWidget(self._window)

        # progress
        self._progress = gui.ProgressBar(self, 100)

        # manage settings
        if self._kWeightSetting != 3:
            self._window._k_spin_box.setValue(self._kWeightSetting)

        # signal / slot connection
        self._window._k_spin_box.valueChanged.connect(self._updateProcess)

    def _updateProcess(self, *arv, **kwargs):
        self._update_settings()
        if self._latest_xas_obj:
            self.process(self._latest_xas_obj)

    @Inputs.data_table
    def processFrmDataTable(self, data_table):
        if data_table is None:
            return
        self.process(Converter.toXASObject(data_table=data_table))

    @Inputs.xas_obj
    def process(self, xas_obj):
        if xas_obj is None:
            return

        if not self._canProcess():
            _logger.warning(
                "There is some processing on going already, will"
                "not process the new dataset"
            )

        self._latest_xas_obj = xas_obj
        self._startProcess()

        # setup the k weight process
        process_obj = QPyMca_k_weight()
        process_obj._advancement.sigProgress.connect(self._setProgressValue)
        process_obj.set_properties(
            {"_kWeightSetting": self._window._k_spin_box.value()}
        )

        # update the processing thread
        thread = self.getProcessingThread()
        thread.init(process_obj=process_obj, xas_obj=self._latest_xas_obj)
        self._callback_finish = functools.partial(
            self._endProcess, self._latest_xas_obj
        )
        thread.finished.connect(self._callback_finish)
        # start processing
        thread.start(priority=qt.QThread.LowPriority)

    def _update_settings(self):
        self._kWeightSetting = self._window._k_spin_box.value()

    def _setProgressValue(self, value):
        self._progress.widget.progressBarSet(value)


class QPyMca_k_weight(est.core.process.pymca.k_weight.PyMca_k_weight):
    """
    Normalization able to give advancement using qt.Signal and QThreadPool
    """

    def __init__(self):
        est.core.process.pymca.k_weight.PyMca_k_weight.__init__(self)
        self._advancement = QProgress("normalization")

    def _pool_process(self, xas_obj):
        self.pool = qt.QThreadPool()
        self.pool.setMaxThreadCount(5)
        for spectrum in xas_obj.spectra:
            assert "KWeight" in xas_obj.configuration
            runnable = ProcessRunnable(
                fct=est.core.process.pymca.k_weight.process_spectr_k,
                spectrum=spectrum,
                configuration=xas_obj.configuration,
                callback=self._advancement.increaseAdvancement,
            )
            self.pool.start(runnable)
        self.pool.waitForDone()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/08/2019"

import unittest
from est.core.types import XASObject
from orangecontrib.est.utils import Converter
from est.core.utils import spectra as spectra_utils
import numpy


class TestConverter(unittest.TestCase):
    """Test conversion of xas_object to Orange.data.Table"""

    def setUp(self):
        self.energy, self.spectra = spectra_utils.create_dataset((128, 20, 1))
        self.xas_object = XASObject(
            energy=self.energy, spectra=self.spectra, dim1=20, dim2=10
        )

    def test_conversion(self):
        """Make sure the conversion to/from Orange.data.Table is safe for energy
        and beam absorption
        """
        xas_object = self.xas_object.copy(create_h5_file=False)
        data_table = Converter.toDataTable(xas_object=xas_object)
        converted_xas_object = Converter.toXASObject(data_table=data_table)
        numpy.testing.assert_array_almost_equal(
            xas_object.energy, converted_xas_object.energy
        )
        numpy.testing.assert_array_almost_equal(
            xas_object.spectra[0].mu, converted_xas_object.spectra[0].mu
        )
        numpy.testing.assert_array_almost_equal(
            xas_object.spectra[5].mu, converted_xas_object.spectra[5].mu
        )
        numpy.testing.assert_array_almost_equal(
            xas_object.spectra[18].mu, converted_xas_object.spectra[18].mu
        )


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestConverter,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite



# coding: utf-8
# ##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "17/05/2017"


import os
import logging
import shutil
from urllib.request import urlopen, ProxyHandler, build_opener

logger = logging.getLogger(__file__)
url_base = "http://www.edna-site.org/pub/est/"


def DownloadDataset(dataset, output_folder, timeout, unpack=False):
    # create if needed path scan
    url = url_base + dataset

    logger.info("Trying to download scan %s, timeout set to %ss", dataset, timeout)
    dictProxies = {}
    if "http_proxy" in os.environ:
        dictProxies["http"] = os.environ["http_proxy"]
        dictProxies["https"] = os.environ["http_proxy"]
    if "https_proxy" in os.environ:
        dictProxies["https"] = os.environ["https_proxy"]
    if dictProxies:
        proxy_handler = ProxyHandler(dictProxies)
        opener = build_opener(proxy_handler).open
    else:
        opener = urlopen
    logger.info("wget %s" % url)
    data = opener(url, data=None, timeout=timeout).read()
    logger.info("Image %s successfully downloaded." % dataset)

    if not os.path.isdir(output_folder):
        os.mkdir(output_folder)

    try:
        archive_folder = os.path.join(output_folder, os.path.basename(dataset))
        with open(archive_folder, "wb") as outfile:
            outfile.write(data)
    except IOError:
        raise IOError(
            "unable to write downloaded \
                        data to disk at %s"
            % archive_folder
        )

    if unpack is True:
        shutil.unpack_archive(archive_folder, extract_dir=output_folder, format="bztar")
        os.remove(archive_folder)


def extract_properties_from_dict(my_str) -> dict:
    """
    Convert parameters provided from a string to a dictionary.
    expected syntax is param1:value1,param2:value2
    would return { param1: value1, param2: value2 }
    try to cast each value to a number (float).
    """
    params = {}
    param_list = my_str.split(",")
    failures = []
    for param in param_list:
        try:
            param_name, param_value = param.split(":")
        except Exception as e:
            logger.info("Fail to cast some parameters: {}".format(e))
            failures.append(param)
        else:
            try:
                param_value = float(param_value)
            except Exception:
                pass
            params[param_name] = param_value
    if len(failures) > 0:
        logger.warning("Fail to convert some parameters : {}".format(failures))
    return params



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tools to visualize spectra"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/07/2019"


from est.core.types import XASObject, Spectrum
from est.core.utils.symbol import MU_CHAR
from silx.gui import qt
from silx.gui.plot import Plot1D
from silx.gui.plot.StackView import StackViewMainWindow
from silx.utils.enum import Enum
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser
from est.gui import icons
from silx.gui import icons as silx_icons
from silx.gui.colors import Colormap
from typing import Iterable
import numpy
import silx
import logging

_logger = logging.getLogger(__name__)

# median spectrum view
silx_version = silx.version.split(".")
if not (int(silx_version[0]) == 0 and int(silx_version[1]) <= 11):
    silx_plot_has_baseline_feature = True
else:
    silx_plot_has_baseline_feature = False
    _logger.warning(
        "a more recent of silx is required to display " "mean spectrum (0.12)"
    )


class ViewType(Enum):
    map = (0,)
    spectrum = (1,)


class _SpectrumViewAction(qt.QAction):
    def __init__(self, parent=None, iView=0):
        qt.QAction.__init__(self, "spectrum view", parent=parent)
        assert iView in (0, 1)  # for now we can only deal with two plot at max
        # otherwise no more icon color to display
        self._iView = iView
        if iView == 0:
            icon = "item-1dim"
        elif iView == 1:
            icon = "item-1dim-black"
        else:
            # if necessary: add more icons, this is the only limitation
            raise NotImplementedError("Only two spectrum views are maanged")
        spectrum_icon = icons.getQIcon(icon)
        self.setIcon(spectrum_icon)
        self.setCheckable(True)


class _MapViewAction(qt.QAction):
    def __init__(self, parent=None):
        qt.QAction.__init__(self, "map view", parent=parent)
        map_icon = silx_icons.getQIcon("image")
        self.setIcon(map_icon)
        self.setCheckable(True)


class XasObjectViewer(qt.QMainWindow):
    """Viewer dedicated to view a XAS object

    :param QObject parent: Qt parent
    :param list mapKeys: list of str keys to propose for the map display
    :param list spectrumsPlots: list of keys if several spectrum plot should be
                                proposed.
    """

    viewTypeChanged = qt.Signal()
    """emitted when the view type change"""

    def __init__(self, parent=None, mapKeys=None, spectrumPlots=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)

        # main stack widget
        self._mainWidget = qt.QWidget(parent=self)
        self._mainWidget.setLayout(qt.QVBoxLayout())
        self.setCentralWidget(self._mainWidget)
        # map view
        self._mapView = MapViewer(parent=self, keys=mapKeys)
        self._mainWidget.layout().addWidget(self._mapView)

        # spectrum view
        self._spectrumViews = []
        if spectrumPlots is not None:
            spectrum_views_ = spectrumPlots
        else:
            spectrum_views_ = ("",)
        for spectrumPlot in range(len(spectrum_views_)):
            spectrumView = SpectrumViewer(parent=self)
            self._mainWidget.layout().addWidget(spectrumView)
            self._spectrumViews.append(spectrumView)
        # add toolbar
        toolbar = qt.QToolBar("")
        toolbar.setIconSize(qt.QSize(32, 32))
        self._spectrumViewActions = []
        self.view_actions = qt.QActionGroup(self)
        for iSpectrumView, tooltip in enumerate(spectrum_views_):
            spectrumViewAction = _SpectrumViewAction(parent=None, iView=iSpectrumView)
            self.view_actions.addAction(spectrumViewAction)
            self._spectrumViewActions.append(spectrumViewAction)
            spectrumViewAction.setToolTip(tooltip)
            toolbar.addAction(spectrumViewAction)
        self._mapViewAction = _MapViewAction()
        toolbar.addAction(self._mapViewAction)
        self.view_actions.addAction(self._mapViewAction)

        self.addToolBar(qt.Qt.LeftToolBarArea, toolbar)
        toolbar.setMovable(False)

        # connect signal / Slot
        self._mapViewAction.triggered.connect(self._updateView)
        for spectrumAction in self._spectrumViewActions:
            spectrumAction.triggered.connect(self._updateView)

        # initialize
        self._spectrumViewActions[0].setChecked(True)
        self._updateView()

    def _updateView(self, *arg, **kwargs):
        index, view_type = self.getViewType()
        self._mapView.setVisible(view_type is ViewType.map)
        for iView, spectrumView in enumerate(self._spectrumViews):
            spectrumView.setVisible(view_type is ViewType.spectrum and iView == index)
        self.viewTypeChanged.emit()

    def getViewType(self):
        if self._mapViewAction.isChecked():
            return None, ViewType.map
        else:
            for spectrumViewAction in self._spectrumViewActions:
                if spectrumViewAction.isChecked():
                    return spectrumViewAction._iView, ViewType.spectrum
        return None, None

    def setXASObj(self, xas_obj):
        self._mapView.clear()

        self._mapView.setXasObject(xas_obj)
        for spectrumView in self._spectrumViews:
            spectrumView.clear()
            spectrumView.setXasObject(xas_obj)


class MapViewer(qt.QWidget):
    """
    Widget to display different map of the spectra
    """

    sigFrameChanged = qt.Signal(int)
    """Signal emitter when the frame number has changed."""

    def __init__(self, parent=None, keys=None):
        """

        :param parent:
        :param keys: volume keys to display for the xasObject (Mu,
        NormalizedMu...)
        """
        assert keys is not None
        self._xasObj = None
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QVBoxLayout())
        self._mainWindow = StackViewMainWindow(parent=parent)
        self.layout().addWidget(self._mainWindow)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0.0)

        self._mainWindow.setKeepDataAspectRatio(True)
        self._mainWindow.setColormap(Colormap(name="temperature"))

        # define the keys combobox
        self._keyWidget = qt.QWidget(parent=self)
        self._keyWidget.setLayout(qt.QHBoxLayout())
        self._keyComboBox = qt.QComboBox(parent=self._keyWidget)
        for key in keys:
            self._keyComboBox.addItem(key)
        self._keyWidget.layout().addWidget(qt.QLabel("view: "))
        self._keyWidget.layout().addWidget(self._keyComboBox)
        self.keySelectionDocker = qt.QDockWidget(parent=self)
        self.keySelectionDocker.setContentsMargins(0, 0, 0, 0)
        self._keyWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._keyWidget.layout().setSpacing(0.0)
        self.keySelectionDocker.setWidget(self._keyWidget)
        # self._mainWindow.addDockWidget(qt.Qt.TopDockWidgetArea, dockWidget)
        self.keySelectionDocker.setAllowedAreas(qt.Qt.TopDockWidgetArea)
        self.keySelectionDocker.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)

        # expose API
        self.getActiveImage = self._mainWindow.getActiveImage
        self.menuBar = self._mainWindow.menuBar

        # connect signal / slot
        self._keyComboBox.currentTextChanged.connect(self._updateView)
        self._mainWindow.sigFrameChanged.connect(self._shareFrameChangedSignal)

    def clear(self):
        self._mainWindow.clear()

    def getActiveKey(self):
        return self._keyComboBox.currentText()

    def setXasObject(self, xas_obj):
        self._xasObj = xas_obj
        self._updateView()

    def _updateView(self, *args, **kwargs):
        if self._xasObj is None:
            return
        # set the map view
        spectra_volume = self._xasObj.spectra.map_to(
            data_info=self.getActiveKey(),
        )
        self._mainWindow.setStack(spectra_volume)

    def getPlot(self):
        return self._mainWindow.getPlot()

    def _shareFrameChangedSignal(self, frame):
        self.sigFrameChanged.emit(frame)

    def setPerspectiveVisible(self, b):
        """hide the dimension selection"""
        self._mainWindow.setOptionVisible(b)
        self._mainWindow._browser.setVisible(True)


class _ExtendedSliderWithBrowser(HorizontalSliderWithBrowser):
    def __init__(self, parent=None, name=None):
        HorizontalSliderWithBrowser.__init__(self, parent)
        self.layout().insertWidget(0, qt.QLabel(str(name + ":")))


class _CurveOperation(object):
    def __init__(
        self,
        x,
        y,
        legend,
        yaxis=None,
        linestyle=None,
        symbol=None,
        color=None,
        ylabel=None,
        baseline=None,
        alpha=1.0,
    ):
        self.x = x
        self.y = y
        self.legend = legend
        self.yaxis = yaxis
        self.linestyle = linestyle
        self.symbol = symbol
        self.color = color
        self.ylabel = ylabel
        self.baseline = baseline
        self.alpha = alpha


class _XMarkerOperation(object):
    def __init__(self, x, legend, color="blue"):
        self.x = x
        self.legend = legend
        self.color = color


class _YMarkerOperation(object):
    def __init__(self, y, legend, color="blue"):
        self.y = y
        self.legend = legend
        self.color = color


class _RawDataList(qt.QTableWidget):
    def __init__(self, parent):
        qt.QTableWidget.__init__(self, parent)
        self.clear()

    def setData(self, x: Iterable, y: Iterable):
        if len(x) != len(y):
            raise ValueError("x and y should have the same number of element")
        self.setRowCount(len(x))
        self.setHorizontalHeaderLabels(
            ["X (energy converted to eV)", "Y ({})".format(MU_CHAR)]
        )

        for i_row, (x_value, y_value) in enumerate(zip(x, y)):
            x_item = qt.QTableWidgetItem()
            x_item.setText(str(x_value))
            self.setItem(i_row, 0, x_item)
            y_item = qt.QTableWidgetItem()
            y_item.setText(str(y_value))
            self.setItem(i_row, 1, y_item)
        self.resizeColumnsToContents()

    def clear(self):
        super().clear()
        self.setHorizontalHeaderLabels(
            ["X (energy converted to eV)", "Y ({})".format(MU_CHAR)]
        )
        self.setRowCount(0)
        self.setColumnCount(2)
        self.setSortingEnabled(True)
        self.verticalHeader().hide()


class SpectrumViewer(qt.QMainWindow):

    sigSpectrumChanged = qt.Signal()
    """Signal emitted when the spectrum change"""

    def __init__(self, parent=None):
        self._curveOperations = []
        """List of callaback to produce plot regarding the XASObject.
        Callback function should return a _curve_operation"""
        qt.QMainWindow.__init__(self, parent)

        self.xas_obj = None

        self._plotWidget = Plot1D(parent=self)
        self._rawDataWidget = _RawDataList(parent=self)

        self._tabWidget = qt.QTabWidget(self)
        self._tabWidget.addTab(self._plotWidget, "plot")
        self._tabWidget.addTab(self._rawDataWidget, "data as text")

        self.setCentralWidget(self._tabWidget)

        # frame browsers
        dockWidget = qt.QDockWidget(self)

        frameBrowsers = qt.QWidget(parent=self)
        frameBrowsers.setLayout(qt.QVBoxLayout())
        frameBrowsers.layout().setContentsMargins(0, 0, 0, 0)

        self._dim1FrameBrowser = _ExtendedSliderWithBrowser(parent=self, name="dim 1")
        frameBrowsers.layout().addWidget(self._dim1FrameBrowser)
        self._dim2FrameBrowser = _ExtendedSliderWithBrowser(parent=self, name="dim 2")
        frameBrowsers.layout().addWidget(self._dim2FrameBrowser)
        dockWidget.setWidget(frameBrowsers)

        self.addDockWidget(qt.Qt.BottomDockWidgetArea, dockWidget)
        dockWidget.setAllowedAreas(qt.Qt.BottomDockWidgetArea)
        dockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)

        # connect signal / slot
        self._dim1FrameBrowser.valueChanged.connect(self._updateSpectrumDisplayed)
        self._dim2FrameBrowser.valueChanged.connect(self._updateSpectrumDisplayed)

    def addCurveOperation(self, callbacks):
        """register an curve to display from Xasobject keys, and a legend

        :param callbacks: callback to call when displaying a specific curve
        :type: Union[list,tuple,function]
        """
        if isinstance(callbacks, (list, tuple)):
            for callback in callbacks:
                self.addCurveOperation(callback)
        else:
            self._curveOperations.append(callbacks)

    def clearCurveOperations(self):
        """Remove all defined curve operation"""
        self._curveOperations.clear()

    def setXasObject(self, xas_obj):
        self.xas_obj = xas_obj
        if self.xas_obj is None:
            self.clear()
        else:
            assert self.xas_obj.spectra.shape[0] >= 0
            assert self.xas_obj.spectra.shape[1] >= 0
            self._dim1FrameBrowser.setRange(0, self.xas_obj.spectra.shape[0] - 1)
            self._dim2FrameBrowser.setRange(0, self.xas_obj.spectra.shape[1] - 1)
            self._updateSpectrumDisplayed()

    def getCurrentSpectrum(self):
        """
        Return the spectrum currently selected
        """
        if self.xas_obj is None:
            return None
        dim1_index = self._dim1FrameBrowser.value()
        dim2_index = self._dim2FrameBrowser.value()
        if dim1_index < 0 or dim2_index < 0:
            return None

        assert dim1_index >= 0
        assert dim2_index >= 0

        return self.xas_obj.get_spectrum(dim1_index, dim2_index)

    def _updateSpectrumDisplayed(self, *args, **kwargs):
        dim1_index = self._dim1FrameBrowser.value()
        spectrum = self.getCurrentSpectrum()
        if spectrum is None:
            return
        # update raw data tab
        self._rawDataWidget.setData(x=spectrum.energy, y=spectrum.mu)
        # update plot tab
        for operation in self._curveOperations:
            curves = operation(spectrum)
            if curves is None:
                continue
            else:
                curves = list(curves)
            if silx_plot_has_baseline_feature is True:
                new_curves_op = operation(self.xas_obj, index=dim1_index)
                if new_curves_op is not None:
                    curves.extend(new_curves_op)
            for res in curves:
                # result can be None or nan if the processing fails. So in this
                # case we won't display anything
                if res is not None:
                    if isinstance(res, _CurveOperation):
                        if res.x is None or res.x is numpy.nan:
                            continue

                        kwargs = {
                            "x": res.x,
                            "y": res.y,
                            "legend": res.legend,
                            "yaxis": res.yaxis,
                            "linestyle": res.linestyle,
                            "symbol": res.symbol,
                            "color": res.color,
                            "ylabel": res.ylabel,
                        }
                        if silx_plot_has_baseline_feature:
                            kwargs["baseline"] = (res.baseline,)

                        curve = self._plotWidget.addCurve(**kwargs)
                        curve = self._plotWidget.getCurve(curve)
                        curve.setAlpha(res.alpha)
                    elif isinstance(res, _XMarkerOperation):
                        if res.x is None or res.x is numpy.nan:
                            continue

                        self._plotWidget.addXMarker(
                            x=res.x, color=res.color, legend=res.legend
                        )
                    elif isinstance(res, _YMarkerOperation):
                        if res.y is None or res.y is numpy.nan:
                            continue

                        self._plotWidget.addYMarker(
                            y=res.y, color=res.color, legend=res.legend
                        )
                    else:
                        raise TypeError(
                            "this type of operation {} is not recognized. Details {}".format(
                                type(res), res
                            )
                        )
        self.sigSpectrumChanged.emit()

    def clear(self):
        self._plotWidget.clear()
        self._dim1FrameBrowser.setMaximum(-1)
        self._dim2FrameBrowser.setMaximum(-1)
        self._rawDataWidget.clear()

    def setYAxisLogarithmic(self, val):
        self._plotWidget.setYAxisLogarithmic(val)


COLOR_MEAN = "black"
COLOR_STD = "grey"


def _get_energy_for(obj, values):
    """Depending on if this is pymca or larch we might need to get
    the normalized energy instead of the energy"""
    if values is None:
        return None
    elif len(values) == len(obj.energy):
        return obj.energy
    else:
        return obj.normalized_energy


def _plot_norm(obj, **kwargs):
    if isinstance(obj, XASObject):
        assert "index" in kwargs
        index_dim1 = kwargs["index"]
        spectra = obj.spectra.map_to("normalized_mu", relative_to="normalized_energy")
        spectra = spectra[:, index_dim1, :]
        mean = numpy.mean(spectra, axis=1)
        std = numpy.std(spectra, axis=1)
        energy = _get_energy_for(obj, mean)
        c1 = _CurveOperation(
            x=energy, y=mean, color=COLOR_MEAN, legend="mean norm", alpha=0.5
        )
        c2 = _CurveOperation(
            x=obj.normalized_energy,
            y=mean + std,
            baseline=mean - std,
            color=COLOR_STD,
            legend="std norm",
            alpha=0.5,
        )
        return c1, c2
    elif isinstance(obj, Spectrum):
        if obj.normalized_mu is None:
            _logger.error("norm has not been computed yet")
            return
        energy = _get_energy_for(obj, obj.normalized_mu)
        c1 = _CurveOperation(
            x=energy, y=obj.normalized_mu, legend="norm", color="black"
        )
        operations = [
            c1,
        ]
        try:
            flat = obj.flat
            operations.append(
                _CurveOperation(
                    x=energy,
                    y=flat,
                    legend="flat normalized mu",
                    color="blue",
                )
            )
        except Exception:
            pass
        return tuple(operations)


def _plot_norm_area(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "norm_area"):
        _logger.error("norm_area has not been computed yet")
        return None
    energy = _get_energy_for(obj, obj.norm_area)

    return (
        _CurveOperation(x=energy, y=obj.norm_area, legend="norm_area", color="orange"),
    )


def _plot_mback_mu(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "mback_mu"):
        _logger.error("mback_mu has not been computed yet")
        return
    energy = _get_energy_for(obj, obj.mback_mu)

    return (
        _CurveOperation(x=energy, y=obj.mback_mu, legend="mback_mu", color="orange"),
    )


def _plot_pre_edge(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if obj.pre_edge is None:
        _logger.error("pre_edge has not been computed yet")
        return
    energy = _get_energy_for(obj, obj.pre_edge)
    return (
        _CurveOperation(x=energy, y=obj.pre_edge, legend="pre edge", color="green"),
    )


def _plot_post_edge(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if obj.post_edge is None:
        _logger.error("post_edge has not been computed yet")
        return
    energy = _get_energy_for(obj, obj.pre_edge)
    return (
        _CurveOperation(x=energy, y=obj.pre_edge, legend="post edge", color="black"),
    )


def _plot_edge(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "e0"):
        _logger.error("e0 has not been computed yet")
        return
    return (_XMarkerOperation(x=obj.e0, legend="edge", color="yellow"),)


def _plot_raw(obj, **kwargs):
    if isinstance(obj, Spectrum):
        if obj.mu is None:
            _logger.error("mu not existing")
            return
        energy = _get_energy_for(obj, obj.mu)
        return (_CurveOperation(x=energy, y=obj.mu, legend="raw mu", color="red"),)
    elif isinstance(obj, XASObject):
        assert "index" in kwargs
        index_dim1 = kwargs["index"]
        spectra = obj.spectra.map_to("normalized_mu", relative_to="normalized_energy")
        spectra = spectra[:, index_dim1, :]
        mean = numpy.mean(spectra, axis=1)
        std = numpy.std(spectra, axis=1)
        energy = _get_energy_for(obj, obj.normalized_energy)

        return (
            _CurveOperation(
                x=energy, y=mean, color=COLOR_MEAN, legend="mean mu", alpha=0.5
            ),
            _CurveOperation(
                x=energy,
                y=mean + std,
                baseline=mean - std,
                color=COLOR_STD,
                legend="mu std",
                alpha=0.5,
            ),
        )
    else:
        raise ValueError("input type is not manged")


def _plot_fpp(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "fpp"):
        _logger.error("fpp has not been computed yet")
        return
    energy = _get_energy_for(obj, obj.fpp)
    return (_CurveOperation(x=energy, y=obj.fpp, legend="fpp", color="blue"),)


def _plot_f2(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "f2"):
        _logger.error("f2 has not been computed yet")
        return
    energy = _get_energy_for(obj, obj.f2)
    return (_CurveOperation(x=energy, y=obj.f2, legend="f2", color="orange"),)


def _plot_chi(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "k"):
        _logger.error("k not computed, unable to display it")
        return
    if not hasattr(obj, "chi"):
        _logger.error("chi not computed, unable to display it")
        return
    return (_CurveOperation(x=obj.k, y=obj.chi, legend="chi(k)"),)


def _plot_chir_mag(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "r"):
        _logger.error("r not computed, unable to display it")
        return
    if not hasattr(obj, "chir_mag"):
        _logger.error("chir_mag not computed, unable to display it")
        return
    return (_CurveOperation(x=obj.r, y=obj.chir_mag, legend="chi k (mag)"),)


def _plot_chir_re(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "r"):
        _logger.error("r not computed, unable to display it")
        return
    if not hasattr(obj, "chir_re"):
        _logger.error("chir_re not computed, unable to display it")
        return
    return (_CurveOperation(x=obj.r, y=obj.chir_re, legend="chi k (real)"),)


def _plot_chir_imag(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "r"):
        _logger.error("r not computed, unable to display it")
        return
    if not hasattr(obj, "chir_im"):
        _logger.error("chir_im not computed, unable to display it")
        return
    return (_CurveOperation(x=obj.r, y=obj.chir_im, legend="chi k (imag)"),)


def _plot_masked_chi_weighted_k(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "masked_k"):
        _logger.error("masked_k not computed, unable to display it")
        return
    if not hasattr(obj, "masked_chi_weighted_k"):
        _logger.error("masked_chi_weighted_k not computed, unable to display it")
        return
    return (
        _CurveOperation(
            x=obj.masked_k, y=obj.masked_chi_weighted_k, legend="chi(k) * k_weight"
        ),
    )


def _plot_spectrum(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    energy = _get_energy_for(obj, obj.mu)
    return (
        _CurveOperation(
            x=energy, y=obj.mu, legend="spectrum", yaxis=None, color="blue"
        ),
    )


def _plot_bkg(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "bkg"):
        _logger.error("missing bkg parameter, unable to compute bkg plot")
        return
    energy = _get_energy_for(obj, obj.bkg)
    return (
        _CurveOperation(x=energy, y=obj.bkg, legend="bkg", yaxis=None, color="red"),
    )


def _plot_knots(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "autobk_details"):
        _logger.error("missing bkg parameter, unable to compute bkg plot")
        return
    if not hasattr(obj.autobk_details, "knots_y"):
        _logger.error("missing knots_y, unable to print output")
        return
    if not hasattr(obj.autobk_details, "knots_e"):
        _logger.error("missing knots_e, unable to print output")
        return

    return (
        _CurveOperation(
            x=obj.autobk_details.knots_e,
            y=obj.autobk_details.knots_y,
            legend="knots",
            yaxis=None,
            color="green",
            linestyle="",
            symbol="o",
        ),
    )


def _exafs_signal_plot(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    missing_keys = obj.get_missing_keys(("EXAFSKValues", "EXAFSSignal"))
    if len(missing_keys) > 0:
        _logger.error(
            "missing keys: {} unable to compute exafs signal plot".format(missing_keys)
        )
        return
    k = obj["EXAFSKValues"]
    if "KMin" not in obj:
        obj["KMin"] = k.min()
    if "KMax" not in obj:
        obj["KMax"] = k.max()

    idx = (obj["EXAFSKValues"] >= obj["KMin"]) & (obj["EXAFSKValues"] <= obj["KMax"])
    x = obj["EXAFSKValues"][idx]
    y = obj["EXAFSSignal"][idx]
    return (_CurveOperation(x=x, y=y, legend="EXAFSSignal"),)


def _exafs_postedge_plot(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    missing_keys = obj.get_missing_keys(("EXAFSKValues", "PostEdgeB"))
    if len(missing_keys) > 0:
        _logger.error(
            "missing keys: {} unable to compute exafs postedge plot".format(
                missing_keys
            )
        )
        return
    k = obj["EXAFSKValues"]
    if "KMin" not in obj:
        obj["KMin"] = k.min()
    if "KMax" not in obj:
        obj["KMax"] = k.max()

    idx = (obj["EXAFSKValues"] >= obj["KMin"]) & (obj["EXAFSKValues"] <= obj["KMax"])

    x = obj["EXAFSKValues"][idx]
    y = obj["PostEdgeB"][idx]
    return (_CurveOperation(x=x, y=y, legend="PostEdge"),)


def _exafs_knots_plot(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    missing_keys = obj.get_missing_keys(("KnotsX", "KnotsY"))
    if len(missing_keys) > 0:
        _logger.error(
            "missing keys: {} unable to compute exafs knot plot".format(missing_keys)
        )
        return
    x = obj["KnotsX"]
    y = obj["KnotsY"]
    return (_CurveOperation(x=x, y=y, legend="Knots", linestyle="", symbol="o"),)


def _normalized_exafs(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    assert isinstance(obj, Spectrum)
    missing_keys = obj.get_missing_keys(("EXAFSKValues", "EXAFSNormalized"))
    if len(missing_keys) > 0:
        _logger.error(
            "missing keys: {} unable to compute normalized EXAFS".format(missing_keys)
        )
        return None

    if obj["KWeight"]:
        if obj["KWeight"] == 1:
            ylabel = "EXAFS Signal * k"
        else:
            ylabel = "EXAFS Signal * k^%d" % obj["KWeight"]
    else:
        ylabel = "EXAFS Signal"

    idx = (obj["EXAFSKValues"] >= obj["KMin"]) & (obj["EXAFSKValues"] <= obj["KMax"])

    return (
        _CurveOperation(
            x=obj["EXAFSKValues"][idx],
            y=obj["EXAFSNormalized"][idx],
            legend="Normalized EXAFS",
            ylabel=ylabel,
        ),
    )


def _ft_window_plot(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    missing_keys = obj.ft.get_missing_keys(("K", "WindowWeight"))
    if len(missing_keys) > 0:
        _logger.error(
            "missing keys {} unable to compute normalized EXAFS".format(missing_keys)
        )
        return None

    return (
        _CurveOperation(
            x=obj.ft["K"],
            y=obj.ft["WindowWeight"],
            legend="FT Window",
            yaxis="right",
            color="red",
        ),
    )


def _ft_intensity_plot(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    missing_keys = obj.ft.get_missing_keys(("FTRadius", "FTIntensity"))
    if len(missing_keys) > 0:
        _logger.error("missing keys:", missing_keys, "unable to compute spectrum plot")
        return
    return (
        _CurveOperation(
            x=obj.ft["FTRadius"], y=obj.ft["FTIntensity"], legend="FT Intensity"
        ),
    )


def _ft_imaginary_plot(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    missing_keys = obj.ft.get_missing_keys(("FTRadius", "FTImaginary"))
    if len(missing_keys) > 0:
        _logger.error("missing keys:", missing_keys, "unable to compute spectrum plot")
        return
    return (
        _CurveOperation(
            x=obj.ft["FTRadius"],
            y=obj.ft["FTImaginary"],
            legend="FT Imaginary",
            color="red",
        ),
    )


def _plot_noise_savgol(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "noise_savgol"):
        _logger.error("noise has not been computed yet")
        return
    energy = obj.noise_savgol_energy
    return (
        _CurveOperation(x=energy, y=obj.noise_savgol, legend="noise", color="blue"),
    )


def _plot_norm_noise_savgol(obj, **kwargs):
    if not isinstance(obj, Spectrum):
        return None
    if not hasattr(obj, "norm_noise_savgol"):
        _logger.error("noise has not been computed yet")
        return
    return (
        _YMarkerOperation(
            y=obj.norm_noise_savgol, legend="normalized noise", color="red"
        ),
    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tools to select energy roi"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "27/01/2021"


from silx.gui import qt
from est.gui.xas_object_definition import _SpectrumPlot


class _RoiSettings(qt.QWidget):

    sigValueChanged = qt.Signal()
    """signal emitted when roi value is changed"""

    MAX_VALUE = 999999999

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QFormLayout())
        self._minE = qt.QDoubleSpinBox(self)
        self._minE.setRange(0, _RoiSettings.MAX_VALUE)
        self._minE.setValue(0)
        self.layout().addRow("min E", self._minE)
        self._maxE = qt.QDoubleSpinBox(self)
        self._maxE.setRange(0, _RoiSettings.MAX_VALUE)
        self._maxE.setValue(_RoiSettings.MAX_VALUE)
        self.layout().addRow("max E", self._maxE)

        # connect signal / slot
        self._minE.editingFinished.connect(self._changed)
        self._maxE.editingFinished.connect(self._changed)

    def _changed(self):
        self.sigValueChanged.emit()

    def setRangeE(self, min_E, max_E):
        old = self.blockSignals(True)
        self._minE.setValue(min_E)
        self._maxE.setValue(max_E)
        self.blockSignals(old)
        self.sigValueChanged.emit()

    def getMinE(self):
        return self._minE.value()

    def setMinE(self, value):
        self._minE.setValue(value)

    def getMaxE(self):
        return self._maxE.value()

    def setMaxE(self, value):
        self._maxE.setValue(value)

    def getROI(self):
        return self.getMinE(), self.getMaxE()

    def setROI(self, roi):
        min, max = roi
        self.setMinE(min)
        self.setMaxE(max)

    def hasntBeenModified(self):
        return (self._minE.value() == 0) and (
            self._maxE.value() == _RoiSettings.MAX_VALUE
        )


class EnergyRoiWidget(qt.QMainWindow):
    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)

        # plot
        self._plot = _SpectrumPlot(self)
        self.setCentralWidget(self._plot)

        # roi settings
        self._widget = _RoiSettings(self)
        dockWidget = qt.QDockWidget(parent=self)
        dockWidget.setWidget(self._widget)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, dockWidget)
        dockWidget.setAllowedAreas(qt.Qt.RightDockWidgetArea | qt.Qt.LeftDockWidgetArea)
        dockWidget.setFeatures(qt.QDockWidget.NoDockWidgetFeatures)

        # add Markers
        self._minEMarker = self._plot.addXMarker(
            self._widget.getMinE(),
            legend="from",
            color="red",
            draggable=True,
            text="min E",
        )

        self._maxEMarker = self._plot.addXMarker(
            self._widget.getMaxE(),
            legend="to",
            color="red",
            draggable=True,
            text="max E",
        )

        # expose API
        self.setROI = self._widget.setROI
        self.getROI = self._widget.getROI

        # connect signal / slot
        self._widget.sigValueChanged.connect(self._updateROIAnchors)
        self._getMinMarker().sigDragFinished.connect(self._updateMinValuefrmMarker)
        self._getMaxMarker().sigDragFinished.connect(self._updateMaxValuefrmMarker)

    def setXasObject(self, xas_obj):
        if xas_obj is None:
            self._plot.clear()
        else:
            self._plot.setXasObject(xas_obj)
            if self._widget.hasntBeenModified() and xas_obj.energy is not None:
                self._widget.setRangeE(xas_obj.energy.min(), xas_obj.energy.max())

    def getXasObject(self):
        return self._plot._plot.xas_obj

    def _getMinMarker(self):
        return self._plot._plot._plotWidget._getMarker(self._minEMarker)

    def _getMaxMarker(self):
        return self._plot._plot._plotWidget._getMarker(self._maxEMarker)

    def _updateROIAnchors(self):
        oldMinMarker = self._getMinMarker().blockSignals(True)
        oldMaxMarker = self._getMaxMarker().blockSignals(True)

        self._getMinMarker().setPosition(self._widget.getMinE(), 0)
        self._getMaxMarker().setPosition(self._widget.getMaxE(), 0)

        self._getMinMarker().blockSignals(oldMinMarker)
        self._getMaxMarker().blockSignals(oldMaxMarker)

    def _updateMinValuefrmMarker(self):
        old = self._widget.blockSignals(True)
        self._widget.setMinE(self._getMinMarker().getPosition()[0])
        self._widget.blockSignals(old)

    def _updateMaxValuefrmMarker(self):
        old = self._widget.blockSignals(True)
        self._widget.setMaxE(self._getMaxMarker().getPosition()[0])
        self._widget.blockSignals(old)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tools to defining auto bk parameters"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/08/2019"


from silx.gui import qt
from est.gui.larch.utils import _OptionalQDoubleSpinBox, _OptionalQIntSpinBox
import logging

_logger = logging.getLogger(__name__)


class _AutobkParameters(qt.QWidget):
    """
    Widget for setting the configuration of the larch 'autobk' process
    """

    sigChanged = qt.Signal()
    """Signal emitted when parameters changed"""
    # TODO: group it with the xftf windows
    _VALID_WINDOWS = ("kaiser", "hanning", "parzen", "welch", "gaussian", "sine")

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QFormLayout())
        # rbkg
        self._rbkgSB = qt.QSpinBox(parent=self)
        self._rbkgSB.setToolTip(
            "distance (in Ang) for chi(R) above which the " "signal is ignored"
        )
        self._rbkgSB.setValue(1)
        self.layout().addRow(qt.QLabel("rbkg"), self._rbkgSB)
        # e0
        self._e0SB = _OptionalQDoubleSpinBox(parent=self)
        self._e0SB.setToolTip(
            "edge energy, in eV.  If None, it will be " "determined here."
        )
        self._e0SB.setMinimum(0.0)
        self._e0SB.setValue(2000.0)
        self._e0SB.setValue(None)
        self.layout().addRow(qt.QLabel("e0", parent=self), self._e0SB)
        # edge step
        self._edgeStepSB = _OptionalQDoubleSpinBox(parent=self)
        self._edgeStepSB.setToolTip("edge step. If None, it will be determined.")
        self._edgeStepSB.setValue(None)
        self.layout().addRow(qt.QLabel("edge step", parent=self), self._edgeStepSB)
        # for now we hide the n knots because it is not managed by larch
        # # n knots
        # self._nknotsSB = _OptionalQIntSpinBox(parent=self)
        # self._nknotsSB.setMinimum(1)
        # self._nknotsSB.setValue(3)
        # self._nknotsSB.setToolTip("number of knots in spline. "
        #                           "If None, it will be determined.")
        # self._nknotsSB.setValue(None)
        # self.layout().addRow(qt.QLabel('nknots', parent=self), self._nknotsSB)
        # kmin
        self._kminSB = qt.QSpinBox(parent=self)
        self._kminSB.setToolTip("minimum k value.")
        self._kminSB.setValue(0)
        self.layout().addRow(qt.QLabel("kmin", parent=self), self._kminSB)
        # kmax
        self._kmaxSB = _OptionalQIntSpinBox(parent=self)
        self._kmaxSB.setToolTip("minimum k value.")
        self._kmaxSB.setValue(None)
        self.layout().addRow(qt.QLabel("kmax", parent=self), self._kmaxSB)
        # kweight
        # TODO: this should be included in the workflow process
        self._kweightSB = qt.QSpinBox(parent=self)
        self._kweightSB.setToolTip("k weight")
        self._kweightSB.setValue(1)
        self.layout().addRow(qt.QLabel("k weight", parent=self), self._kweightSB)
        # dk
        self._dkSB = qt.QDoubleSpinBox(parent=self)
        self._dkSB.setMinimum(0)
        self._dkSB.setValue(0.1)
        self._dkSB.setToolTip("FFT window window parameter")
        self.layout().addRow(qt.QLabel("FFT window parameter", parent=self), self._dkSB)
        # window name
        self._windowCB = qt.QComboBox(parent=self)
        for winName in self._VALID_WINDOWS:
            self._windowCB.addItem(winName)
        index = self._windowCB.findText("hanning")
        assert index >= 0
        self._windowCB.setCurrentIndex(index)
        self.layout().addRow(qt.QLabel("fft window type", parent=self), self._windowCB)
        # nfft
        self._nfftSB = qt.QSpinBox(parent=self)
        self._nfftSB.setMaximum(999999)
        self._nfftSB.setValue(2048)
        self._nfftSB.setToolTip("value to use for N_fft")
        self.layout().addRow(qt.QLabel("N fft", parent=self), self._nfftSB)
        # kstep
        self._kstepSB = qt.QDoubleSpinBox(parent=self)
        self._kstepSB.setToolTip("value to use for delta_k")
        self._kstepSB.setValue(0.05)
        self._kstepSB.setSingleStep(0.05)
        self.layout().addRow(qt.QLabel("k step", parent=self), self._kstepSB)
        # nclamp
        self._nclampSB = qt.QSpinBox(parent=self)
        self._nclampSB.setValue(4)
        self._nclampSB.setToolTip("number of energy end-points for clamp")
        self.layout().addRow(qt.QLabel("n clamp", parent=self), self._nclampSB)
        # clamp_lo
        self._clampLoSB = qt.QSpinBox(parent=self)
        self._clampLoSB.setValue(1)
        self._clampLoSB.setToolTip("weight of low-energy clamp")
        self.layout().addRow(qt.QLabel("clamp low", parent=self), self._clampLoSB)
        # clamp_hi
        self._clampHiSB = qt.QSpinBox(parent=self)
        self._clampHiSB.setValue(1)
        self._clampHiSB.setToolTip("weight of high-energy clamp")
        self.layout().addRow(qt.QLabel("clamp high", parent=self), self._clampHiSB)
        # calc uncertainties
        self._calcUncertaintiesCB = qt.QCheckBox(parent=self)
        self._calcUncertaintiesCB.setChecked(True)
        self._calcUncertaintiesCB.setToolTip(
            "alculate uncertainties in " "mu_0(E) and chi(k)"
        )
        self.layout().addRow(
            qt.QLabel("calculate uncertainties", parent=self), self._calcUncertaintiesCB
        )
        # err sigma
        self._errSignalSB = qt.QSpinBox(parent=self)
        self._errSignalSB.setValue(1)
        self.layout().addRow(
            qt.QLabel("sigma uncertainties level", parent=self), self._errSignalSB
        )

        # expose API
        self.getRbkg = self._rbkgSB.value
        self.setRbkg = self._rbkgSB.setValue
        self.getE0 = self._e0SB.getValue
        self.setE0 = self._e0SB.setValue
        self.getEdgeStep = self._edgeStepSB.getValue
        self.setEdgeStep = self._edgeStepSB.setValue
        # self.getKNots = self._nknotsSB.getValue
        # self.setKNots = self._nknotsSB.setValue
        self.getKMin = self._kminSB.value
        self.setKMin = self._kminSB.setValue
        self.getKMax = self._kmaxSB.getValue
        self.setKMax = self._kmaxSB.setValue
        self.getKWeight = self._kweightSB.value
        self.setKWeight = self._kweightSB.setValue
        self.getDk = self._dkSB.value
        self.setDk = self._dkSB.setValue
        self.getWindow = self._windowCB.currentText
        self.setWindow = self.setWindow
        self.getNfft = self._nfftSB.value
        self.setNfft = self._nfftSB.setValue
        self.getKStep = self._kstepSB.value
        self.setKStep = self._kstepSB.setValue
        self.getNClamp = self._nclampSB.value
        self.setNClamp = self._nclampSB.setValue
        self.getClampLow = self._clampLoSB.value
        self.setClampLow = self._clampLoSB.setValue
        self.getClampHigh = self._clampHiSB.value
        self.setClampHigh = self._clampHiSB.setValue
        self.isCalcUncertainties = self._calcUncertaintiesCB.isChecked
        self.setCalcUncertainties = self._calcUncertaintiesCB.setChecked
        self.getSigmaErr = self._errSignalSB.value
        self.setSigmaErr = self._errSignalSB.setValue

        # connect signal / slots
        self._rbkgSB.valueChanged.connect(self._valueChanged)
        self._e0SB.sigChanged.connect(self._valueChanged)
        self._edgeStepSB.sigChanged.connect(self._valueChanged)
        # self._nknotsSB.sigChanged.connect(self._parametersChanged)
        self._kminSB.valueChanged.connect(self._valueChanged)
        self._kmaxSB.sigChanged.connect(self._valueChanged)
        self._kweightSB.valueChanged.connect(self._valueChanged)
        self._dkSB.valueChanged.connect(self._valueChanged)
        self._windowCB.currentTextChanged.connect(self._valueChanged)
        self._nfftSB.valueChanged.connect(self._valueChanged)
        self._kstepSB.valueChanged.connect(self._valueChanged)
        self._nclampSB.valueChanged.connect(self._valueChanged)
        self._clampLoSB.valueChanged.connect(self._valueChanged)
        self._clampHiSB.valueChanged.connect(self._valueChanged)
        self._calcUncertaintiesCB.toggled.connect(self._valueChanged)
        self._errSignalSB.valueChanged.connect(self._valueChanged)

        # avoid several emission of the sigChanged signal
        self._lastParameters = None

    def getParameters(self):
        return {
            "rbkg": self.getRbkg(),
            "e0": self.getE0(),
            "edge_step": self.getEdgeStep(),
            # 'nknots': self.getKNots(),
            "kmin": self.getKMin(),
            "kmax": self.getKMax(),
            "kweight": self.getKWeight(),
            "dk": self.getDk(),
            "win": self.getWindow(),
            "nfft": self.getNfft(),
            "kstep": self.getKStep(),
            "nclamp": self.getNClamp(),
            "clamp_lo": self.getClampLow(),
            "clamp_hi": self.getClampHigh(),
            "calc_uncertaintites": self.isCalcUncertainties(),
            "err_sigma": self.getSigmaErr(),
        }

    def setParameters(self, parameters):
        assert isinstance(parameters, dict)
        for key, value in parameters.items():
            if key == "rbkg":
                self.setRbkg(value)
            elif key == "e0":
                self.setE0(value)
            elif key == "edge_step":
                self.setEdgeStep(value)
            elif key == "knots":
                self.setKNots(value)
            elif key == "kmin":
                self.setKMin(value)
            elif key == "kmax":
                self.setKMax(value)
            elif key == "kweight":
                self.setKWeight(value)
            elif key == "dk":
                self.setDk(value)
            elif key == "win":
                self.setWindow(value)
            elif key == "nfft":
                self.setNfft(value)
            elif key == "kstep":
                self.setKStep(value)
            elif key == "nclamp":
                self.setNClamp(value)
            elif key == "clamp_lo":
                self.setClampLow(value)
            elif key == "clamp_hi":
                self.setClampHigh(value)
            elif key == "calc_uncertaintites":
                self.setCalcUncertainties(value)
            elif key == "err_sigma":
                self.setSigmaErr(value)

    def setWindow(self, window):
        assert window in self._VALID_WINDOWS
        index = self._windowCB.findText(window)
        if index >= 0:
            self._windowCB.setCurrentIndex(index)
        else:
            _logger.warning(window + " not found.")

    def _valueChanged(self, *args, **kwargs):
        currentParameters = self.getParameters()
        if currentParameters != self._lastParameters:
            self._lastParameters = currentParameters
            self.sigChanged.emit()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""wrapper to the larch autobk process"""

from est.core.types import Spectrum, XASObject
from est.core.process.process import Process
from est.core.process.process import _NexusDatasetDef
from est.core.process.process import _input_desc
from est.core.process.process import _output_desc
from larch.xafs.autobk import autobk
import multiprocessing
import functools
import logging

_logger = logging.getLogger(__name__)

_DEBUG = True
if _DEBUG:
    from larch.symboltable import Group


def process_spectr_autobk(
    spectrum,
    configuration,
    overwrite=True,
    callbacks=None,
    output=None,
    output_dict=None,
):
    """

    :param spectrum: spectrum to process
    :type: :class:`.Spectrum`
    :param configuration: configuration of the pymca normalization
    :type: dict
    :param overwrite: False if we want to return a new Spectrum instance
    :type: bool
    :param callback: callback to execute.
    :param output: list to store the result, needed for pool processing
    :type: multiprocessing.manager.list
    :param output_dict: key is input spectrum, value is index in the output
                        list.
    :type: dict
    :return: processed spectrum
    :rtype: tuple (configuration, spectrum)
    """
    _logger.debug("start autobk on spectrum (%s, %s)" % (spectrum.x, spectrum.y))
    assert isinstance(spectrum, Spectrum)
    if spectrum.energy is None or spectrum.mu is None:
        _logger.error("Energy and or Mu is/are not specified, unable to compute autpbk")
        return None, None

    _conf = configuration
    if "autobk" in _conf:
        _conf = _conf["autobk"]
    opts = {}
    for opt_name in (
        "rbkg",
        "nknots",
        "e0",
        "edge_step",
        "kmin",
        "kmax",
        "kweight",
        "dk",
        "win",
        "k_std",
        "chi_std",
        "nfft",
        "kstep",
        "pre_edge_kws",
        "nclamp",
        "clamp_lo",
        "clamp_hi",
        "calc_uncertainties",
        "err_sigma",
    ):
        if opt_name in _conf:
            opts[opt_name] = _conf[opt_name]
    if "pre_edge" in configuration:
        _conf["pre_edge_kws"] = configuration["pre_edge"]

    if _DEBUG is True:
        assert isinstance(spectrum, Group)
    if overwrite:
        _spectrum = spectrum
    else:
        _spectrum = Spectrum().load_frm_dict(spectrum.to_dict())
    try:
        autobk(_spectrum, group=_spectrum, **opts)
    except ValueError as e:
        _logger.error(e)
    if callbacks:
        for callback in callbacks:
            callback()
    return configuration, _spectrum


def larch_autobk(xas_obj):
    """

    :param xas_obj: object containing the configuration and spectra to process
    :type: Union[XASObject, dict]
    :return: spectra dict
    :rtype: XASObject
    """
    mback_obj = Larch_autobk()
    return mback_obj.process(xas_obj=xas_obj)


_USE_MULTIPROCESSING_POOL = False
# note: we cannot use multiprocessing pool with pypushflow for now.


class Larch_autobk(Process):

    inputs = [
        _input_desc(name="xas_obj", type=XASObject, handler="process", doc=""),
    ]

    outputs = [
        _output_desc(name="xas_obj", type=XASObject, doc=""),
    ]

    def __init__(self):
        Process.__init__(self, name="autobk")

    def set_properties(self, properties):
        if "_larchSettings" in properties:
            self._settings = properties["_larchSettings"]

    def process(self, xas_obj):
        _xas_obj = self.getXasObject(xas_obj=xas_obj)
        if self._settings:
            _xas_obj.configuration["autobk"] = self._settings

        self._advancement.reset(max_=_xas_obj.n_spectrum)
        self._advancement.startProcess()
        self._pool_process(xas_obj=_xas_obj)
        self._advancement.endProcess()
        self.register_process(_xas_obj, data_keys=(_NexusDatasetDef("bkg"),))
        return _xas_obj

    def _pool_process(self, xas_obj):
        assert isinstance(xas_obj, XASObject)
        if not _USE_MULTIPROCESSING_POOL:
            for spectrum in xas_obj.spectra:
                process_spectr_autobk(
                    spectrum=spectrum,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=True,
                )
        else:
            from multiprocessing import Manager

            manager = Manager()
            output_dict = {}
            res_list = manager.list()
            for i_spect, spect in enumerate(xas_obj.spectra):
                res_list.append(None)
                output_dict[spect] = i_spect

            with multiprocessing.Pool(5) as p:
                partial_ = functools.partial(
                    process_spectr_autobk,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=False,
                    output=res_list,
                    output_dict=output_dict,
                )
                p.map(partial_, xas_obj.spectra)

            # then update local spectrum
            for spectrum, res in zip(xas_obj.spectra, res_list):
                spectrum.update(res)

    def definition(self):
        return "autobk calculation"

    def program_version(self):
        import larch.version

        return larch.version.version_data()["larch"]

    def program_name(self):
        return "larch_autobk"

    __call__ = process



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/26/2019"


import unittest
from est.core.types import Dim
from est.core.io import XASReader
from silx.io.url import DataUrl
from est.core.types import XASObject
from est.io.utils.information import InputInformation
from est.units import ur
import numpy
import os
import tempfile
import h5py
import shutil


class TestSpectraDimensions(unittest.TestCase):
    """
    Test reading spectra with different dimensions organisation
    (X, Y, Channels), (Channels, Y, X), (Y, Channels, X)
    """

    def setUp(self) -> None:
        self.spectra_path = "/data/NXdata/data"
        self.channel_path = "/data/NXdata/Channel"
        self.output_dir = tempfile.mkdtemp()
        self.filename = os.path.join(self.output_dir, "myfile.h5")

    def tearDown(self) -> None:
        shutil.rmtree(self.output_dir)

    def saveSpectra(self, spectra):
        """Save the spectra to the spectra file defined in setup and return the
        associated silx url"""
        with h5py.File(self.filename, "a") as f:
            f[self.spectra_path] = spectra

        return DataUrl(
            file_path=self.filename, data_path=self.spectra_path, scheme="silx"
        )

    def saveChannel(self, channel):
        """Save the energy to the spectra file defined in setup and return the
        associated silx url"""
        with h5py.File(self.filename, "a") as f:
            f[self.channel_path] = channel

        return DataUrl(
            file_path=self.filename, data_path=self.channel_path, scheme="silx"
        )

    def testDimensionsXYEnergy(self):
        """Test that spectra stored as X, Y Energy can be read"""
        x_dim = 4
        y_dim = 2
        energy_dim = 3
        shape = (x_dim, y_dim, energy_dim)
        spectra = numpy.arange(x_dim * y_dim * energy_dim).reshape(shape)
        channel = numpy.linspace(0, 1, energy_dim)
        spectra_url = self.saveSpectra(spectra)
        channel_url = self.saveChannel(channel=channel)

        # if dims are incoherent with energy, should raise an error
        dims = (Dim.DIM_2, Dim.DIM_1, Dim.DIM_0)
        with self.assertRaises(ValueError):
            XASReader().read_frm_url(
                InputInformation(
                    spectra_url=spectra_url,
                    channel_url=channel_url,
                    dimensions=dims,
                )
            )
        dims = (Dim.DIM_0, Dim.DIM_1, Dim.DIM_2)
        xas_obj = XASReader().read_frm_url(
            InputInformation(
                spectra_url=spectra_url,
                channel_url=channel_url,
                dimensions=dims,
            )
        )
        self.assertTrue(isinstance(xas_obj, XASObject))
        self.assertTrue(xas_obj.n_spectrum == x_dim * y_dim)
        numpy.testing.assert_array_equal(
            xas_obj.spectra.data.flat[1].mu, spectra[1, 0, :]
        )
        numpy.testing.assert_array_equal(xas_obj.spectra.data.flat[2].energy, channel)

    def testDimensionsChannelYX(self):
        """Test that spectra stored as Channel, Y, X can be read"""
        x_dim = 10
        y_dim = 5
        energy_dim = 30
        shape = (energy_dim, y_dim, x_dim)
        spectra = numpy.arange(x_dim * y_dim * energy_dim).reshape(shape)
        channel = numpy.linspace(0, 100, energy_dim)
        spectra_url = self.saveSpectra(spectra)
        channel_url = self.saveChannel(channel=channel)

        # if dims are incoherent with energy, should raise an error
        dims = (Dim.DIM_0, Dim.DIM_1, Dim.DIM_2)
        with self.assertRaises(ValueError):
            XASReader().read_frm_url(
                InputInformation(
                    spectra_url=spectra_url,
                    channel_url=channel_url,
                    dimensions=dims,
                )
            )

        dims = (Dim.DIM_2, Dim.DIM_1, Dim.DIM_0)
        xas_obj = XASReader().read_frm_url(
            InputInformation(
                spectra_url=spectra_url, channel_url=channel_url, dimensions=dims
            )
        )
        self.assertTrue(isinstance(xas_obj, XASObject))
        self.assertTrue(xas_obj.n_spectrum == x_dim * y_dim)
        numpy.testing.assert_array_equal(
            xas_obj.spectra.data.flat[1].mu, spectra[:, 0, 1]
        )
        numpy.testing.assert_array_equal(
            xas_obj.spectra.data.flat[2].energy, (channel * ur.eV).m
        )


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestSpectraDimensions,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")
